# ble_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""BLE API for Open GoPro"""

from __future__ import annotations
import logging
import datetime
from typing import Optional, Final, Callable

from construct import (
    Int8ub,
    Int32ub,
    Int64ub,
    Flag,
    GreedyBytes,
    GreedyString,
    Struct,
    Padding,
    PaddedString,
    this,
    Hex,
    Int16ub,
    Int16sb,
)

from open_gopro import proto
from open_gopro.responses import GoProResp, BytesParser, BytesBuilder, JsonParser
from open_gopro.interface import GoProBle, Commands, BleCommand
from open_gopro.api.builders import (
    DeprecatedAdapter,
    BleStatus,
    BleSetting,
    BleAsyncResponse,
    BleWriteCommand,
    RegisterUnregisterAll,
    BleReadCommand,
    BleProtoCommand,
)
from open_gopro.constants import ActionId, FeatureId, CmdId, QueryCmdId, SettingId, StatusId, GoProUUIDs
from open_gopro.util import map_keys
from . import params as Params

logger = logging.getLogger(__name__)


class BleCommands(Commands[BleCommand, CmdId]):
    """All of the BLE commands.

    To be used as a delegate for a GoProBle instance to build commands

    All of these return a GoProResp
    """

    # pylint: disable = missing-class-docstring, arguments-differ, useless-super-delegation
    def __init__(self, communicator: GoProBle):
        """Constructor

        Args:
            communicator (GoProBle): GoPro BLE communicator that will send the commands
        """
        ######################################################################################################
        #                          BLE WRITE COMMANDS
        ######################################################################################################

        class SetShutter(BleWriteCommand):
            def __call__(self, shutter: Params.Toggle) -> GoProResp:
                """Set the Shutter

                Args:
                    shutter (open_gopro.api.params.Toggle): on or off

                Returns:
                    GoProResp: status of command
                """
                return super().__call__(shutter=shutter)

        #: Sphinx docstring redirect
        self.set_shutter = SetShutter(communicator, GoProUUIDs.CQ_COMMAND, CmdId.SET_SHUTTER, Int8ub)

        class TagHilight(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Tag a highlight during encoding

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.tag_hilight = TagHilight(communicator, GoProUUIDs.CQ_COMMAND, CmdId.TAG_HILIGHT)

        class PowerDown(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Power Down the camera

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.power_down = PowerDown(communicator, GoProUUIDs.CQ_COMMAND, CmdId.POWER_DOWN)

        class Sleep(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Put the camera in standby

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.sleep = Sleep(communicator, GoProUUIDs.CQ_COMMAND, CmdId.SLEEP)

        class EnableWifi(BleWriteCommand):
            def __call__(self, enable: bool) -> GoProResp:
                """Enable / disable the Wi-Fi Access Point.

                Args:
                    enable (bool): True to enable, False to disable

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__(enable=enable)

        #: Sphinx docstring redirect
        self.enable_wifi_ap = EnableWifi(communicator, GoProUUIDs.CQ_COMMAND, CmdId.SET_WIFI, Int8ub)

        class GetHardwareInfo(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get the model number, board, type, firmware version, serial number, and AP info

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_hardware_info = GetHardwareInfo(
            communicator,
            GoProUUIDs.CQ_COMMAND,
            CmdId.GET_HW_INFO,
            parser=Struct(
                Padding(1),
                "model_number" / Int32ub,
                "model_name_len" / Int8ub,
                "model_name" / PaddedString(this.model_name_len, "utf-8"),
                Padding(1),
                "board_type" / Hex(Int32ub),
                "firmware_version_len" / Int8ub,
                "firmware_version" / PaddedString(this.firmware_version_len, "utf-8"),
                "serial_number_len" / Int8ub,
                "serial_number" / PaddedString(this.serial_number_len, "utf-8"),
                "ap_ssid_len" / Int8ub,
                "ap_ssid" / PaddedString(this.ap_ssid_len, "utf-8"),
                "ap_mac_len" / Int8ub,
                "ap_mac" / PaddedString(this.ap_mac_len, "utf-8"),
            ),
        )

        class LoadPresetGroup(BleWriteCommand):
            def __call__(self, group: Params.PresetGroup) -> GoProResp:
                """Load a Preset Group.

                Once complete, the most recently used preset in this group will be active.


                Args:
                    group (open_gopro.api.params.PresetGroup): preset group to load

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__(group=group)

        #: Sphinx docstring redirect
        self.load_preset_group = LoadPresetGroup(
            communicator, GoProUUIDs.CQ_COMMAND, CmdId.LOAD_PRESET_GROUP, Int16ub
        )

        class LoadPreset(BleWriteCommand):
            def __call__(self, preset: int) -> GoProResp:
                """Load a Preset

                The int ID value can be found from the get_preset_status command

                Args:
                    preset (int): preset ID to load

                Returns:
                    GoProResp: command status
                """
                return super().__call__(preset=preset)

        #: Sphinx docstring redirect
        self.load_preset = LoadPreset(communicator, GoProUUIDs.CQ_COMMAND, CmdId.LOAD_PRESET, Int32ub)

        class SetThirdPartyClientInfo(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Flag as third party app

                Returns:
                    GoProResp: command status
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.set_third_party_client_info = SetThirdPartyClientInfo(
            communicator, GoProUUIDs.CQ_COMMAND, CmdId.SET_THIRD_PARTY_CLIENT_INFO
        )

        class GetOgpApiVersion(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get Open GoPro API Version

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_open_gopro_api_version = GetOgpApiVersion(
            communicator,
            GoProUUIDs.CQ_COMMAND,
            CmdId.GET_THIRD_PARTY_API_VERSION,
            parser=Struct(Padding(1), "major" / Int8ub, Padding(1), "minor" / Int8ub),
        )

        class GetCameraStatuses(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get all of the camera's statuses

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_camera_statuses = GetCameraStatuses(
            communicator, GoProUUIDs.CQ_QUERY, CmdId.GET_CAMERA_STATUSES
        )

        class GetCameraSettings(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get all of the camera's settings

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_camera_settings = GetCameraSettings(
            communicator, GoProUUIDs.CQ_QUERY, CmdId.GET_CAMERA_SETTINGS
        )

        class GetCameraCapabilities(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get the current capabilities of each camera setting

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_camera_capabilities = GetCameraCapabilities(
            communicator, GoProUUIDs.CQ_QUERY, CmdId.GET_CAMERA_CAPABILITIES
        )

        class SetDateTime(BleWriteCommand):
            def __call__(self, date_time: datetime.datetime) -> GoProResp:
                """Set the camera's date and time (non timezone / DST version)

                Args:
                    date_time (datetime.datetime): Date and time to set (Timezone will be ignored)

                Returns:
                    GoProResp: command status
                """
                return super().__call__(date_time=date_time)

        #: Sphinx docstring redirect
        self.set_date_time = SetDateTime(
            communicator,
            GoProUUIDs.CQ_COMMAND,
            CmdId.SET_DATE_TIME,
            param_builder=BleParserBuilders.DateTime(),
        )

        class GetDateTime(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get the camera's date and time (non timezone / DST version)

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_date_time = GetDateTime(
            communicator,
            GoProUUIDs.CQ_COMMAND,
            CmdId.GET_DATE_TIME,
            parser=BleParserBuilders.DateTime(),
        )

        class SetDateTimeTzDst(BleWriteCommand):
            def __call__(self, date_time: datetime.datetime, tz_offset: int, is_dst: bool) -> GoProResp:
                """Set the camera's date and time with timezone and DST

                Args:
                    date_time (datetime.datetime): date and time
                    tz_offset (int): timezone as UTC offset
                    is_dst (bool): is daylight savings time?

                Returns:
                    GoProResp: command status
                """
                return super().__call__(date_time=date_time, tz_offset=tz_offset, is_dst=is_dst)

        #: Sphinx docstring redirect
        self.set_date_time_tz_dst = SetDateTimeTzDst(
            communicator,
            GoProUUIDs.CQ_COMMAND,
            CmdId.SET_DATE_TIME_DST,
            param_builder=BleParserBuilders.DateTime(),
        )

        class GetDateTimeTzDst(BleWriteCommand):
            def __call__(self) -> GoProResp:
                """Get the camera's date and time with timezone / DST

                Returns:
                    GoProResp: response as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_date_time_tz_dst = GetDateTimeTzDst(
            communicator,
            GoProUUIDs.CQ_COMMAND,
            CmdId.GET_DATE_TIME_DST,
            parser=BleParserBuilders.DateTime(),
        )

        ######################################################################################################
        #                          BLE DIRECT CHARACTERISTIC READ COMMANDS
        ######################################################################################################

        #: Sphinx docstring redirect
        self.get_wifi_ssid = BleReadCommand(
            communicator, GoProUUIDs.WAP_SSID, Struct("ssid" / GreedyString("utf-8"))
        )
        """Get the Wifi SSID.

        Returns:
            GoProResp: command status and SSID
        """

        #: Sphinx docstring redirect
        self.get_wifi_password = BleReadCommand(
            communicator, GoProUUIDs.WAP_PASSWORD, Struct("password" / GreedyString("utf-8"))
        )
        """Get the Wifi password.

        Returns:
            GoProResp: command status and password
        """

        ######################################################################################################
        #                          REGISTER / UNREGISTER ALL COMMANDS
        ######################################################################################################

        self.register_for_all_statuses = RegisterUnregisterAll(
            communicator,
            GoProUUIDs.CQ_QUERY,
            CmdId.REGISTER_ALL_STATUSES,
            producer=(StatusId, QueryCmdId.STATUS_VAL_PUSH),
            action=RegisterUnregisterAll.Action.REGISTER,
        )
        """Register push notifications for all statuses

        Returns:
            GoProResp: command status and current value of all statuses
        """

        self.unregister_for_all_statuses = RegisterUnregisterAll(
            communicator,
            GoProUUIDs.CQ_QUERY,
            CmdId.UNREGISTER_ALL_STATUSES,
            producer=(StatusId, QueryCmdId.STATUS_VAL_PUSH),
            action=RegisterUnregisterAll.Action.UNREGISTER,
        )
        """Unregister push notifications for all statuses

        Returns:
            GoProResp: command status
        """

        self.register_for_all_settings = RegisterUnregisterAll(
            communicator,
            GoProUUIDs.CQ_QUERY,
            CmdId.REGISTER_ALL_SETTINGS,
            producer=(SettingId, QueryCmdId.SETTING_VAL_PUSH),
            action=RegisterUnregisterAll.Action.REGISTER,
        )
        """Register push notifications for all settings

        Returns:
            GoProResp: command status and current value of all settings
        """

        self.unregister_for_all_settings = RegisterUnregisterAll(
            communicator,
            GoProUUIDs.CQ_QUERY,
            CmdId.UNREGISTER_ALL_SETTINGS,
            producer=(SettingId, QueryCmdId.SETTING_VAL_PUSH),
            action=RegisterUnregisterAll.Action.UNREGISTER,
        )
        """Unregister push notifications for all settings

        Returns:
            GoProResp: command status
        """

        self.register_for_all_capabilities = RegisterUnregisterAll(
            communicator,
            GoProUUIDs.CQ_QUERY,
            CmdId.REGISTER_ALL_CAPABILITIES,
            producer=(SettingId, QueryCmdId.SETTING_CAPABILITY_PUSH),
            action=RegisterUnregisterAll.Action.REGISTER,
        )
        """Register push notifications for all capabilities

        Returns:
            GoProResp: command status and current value of all capabilities
        """

        self.unregister_for_all_capabilities = RegisterUnregisterAll(
            communicator,
            GoProUUIDs.CQ_QUERY,
            CmdId.UNREGISTER_ALL_CAPABILITIES,
            producer=(SettingId, QueryCmdId.SETTING_CAPABILITY_PUSH),
            action=RegisterUnregisterAll.Action.UNREGISTER,
        )
        """Unregister push notifications for all capabilities

        Returns:
            GoProResp: command status
        """

        ######################################################################################################
        #                          PROTOBUF COMMANDS
        ######################################################################################################

        class SetCameraControl(BleProtoCommand):
            def __call__(self, control: Params.CameraControlStatus) -> GoProResp:
                """Tell the camera that the app (i.e. External Control) wishes to claim control of the camera.

                Args:
                    control (bool): True to enable, False to disable.

                Returns:
                    GoProResp: command status of request
                """
                return super().__call__(camera_control_status=control)

        #: Sphinx docstring redirect
        self.set_camera_control = SetCameraControl(
            communicator,
            uuid=GoProUUIDs.CQ_COMMAND,
            feature_id=FeatureId.COMMAND,
            action_id=ActionId.SET_CAMERA_CONTROL,
            response_action_id=ActionId.SET_CAMERA_CONTROL_RSP,
            request_proto=proto.RequestSetCameraControlStatus,
            response_proto=proto.ResponseGeneric,
        )

        class TurboMode(BleProtoCommand):
            def __call__(self, active: bool) -> GoProResp:
                """Enable / disable turbo mode.

                Args:
                    active (bool): True to enable, False to disable.

                Returns:
                    GoProResp: command status of request
                """
                return super().__call__(active=active)

        #: Sphinx docstring redirect
        self.set_turbo_mode = TurboMode(
            communicator,
            uuid=GoProUUIDs.CQ_COMMAND,
            feature_id=FeatureId.COMMAND,
            action_id=ActionId.SET_TURBO_MODE,
            response_action_id=ActionId.SET_TURBO_MODE_RSP,
            request_proto=proto.RequestSetTurboActive,
            response_proto=proto.ResponseGeneric,
        )

        class GetPresetStatus(BleProtoCommand):
            def __call__(
                self,
                register_preset_status: Optional[list[Params.RegisterPreset]] = None,
                unregister_preset_status: Optional[list[Params.RegisterPreset]] = None,
            ) -> GoProResp:
                """Get information about what Preset Groups and Presets the camera supports in its current state

                Also optionally (un)register for preset / group preset modified notifications which  will be
                sent asynchronously as :py:attr:`open_gopro.constants.ActionId.PRESET_MODIFIED_NOTIFICATION`

                Args:
                    register_preset_status (list[open_gopro.api.params.RegisterPreset], Optional): Types of
                        preset modified updates to register for. Defaults to None.
                    unregister_preset_status (list[open_gopro.api.params.RegisterPreset], Optional): Types of
                        preset modified updates to unregister for. Defaults to None.

                Returns:
                    GoProResp: JSON data describing all currently available presets
                """
                return super().__call__(
                    register_preset_status=register_preset_status or [],
                    unregister_preset_status=unregister_preset_status or [],
                )

        #: Sphinx docstring redirect
        self.get_preset_status = GetPresetStatus(
            communicator,
            uuid=GoProUUIDs.CQ_QUERY,
            feature_id=FeatureId.QUERY,
            action_id=ActionId.GET_PRESET_STATUS,
            response_action_id=ActionId.GET_PRESET_STATUS_RSP,
            request_proto=proto.RequestGetPresetStatus,
            response_proto=proto.NotifyPresetStatus,
            additional_matching_ids={ActionId.PRESET_MODIFIED_NOTIFICATION},
        )

        class ScanWifiNetworks(BleProtoCommand):
            def __call__(self) -> GoProResp:
                """Scan for Wifi networks

                Returns:
                    GoProResp: Command status of request
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.scan_wifi_networks = ScanWifiNetworks(
            communicator,
            uuid=GoProUUIDs.CM_NET_MGMT_COMM,
            feature_id=FeatureId.NETWORK_MANAGEMENT,
            action_id=ActionId.SCAN_WIFI_NETWORKS,
            response_action_id=ActionId.SCAN_WIFI_NETWORKS_RSP,
            request_proto=proto.RequestStartScan,
            response_proto=proto.ResponseStartScanning,
        )

        class GetAPEntries(BleProtoCommand):
            def __call__(self, scan_id: int) -> GoProResp:
                """Get the results of a scan for wifi networks

                Args:
                    scan_id (int): TODO

                Returns:
                    GoProResp: result of scan with entries for WiFi networks
                """
                return super().__call__(scan_id=scan_id, start_index=0, max_entries=100)

        #: Sphinx docstring redirect
        self.get_ap_entries = GetAPEntries(
            communicator,
            uuid=GoProUUIDs.CM_NET_MGMT_COMM,
            feature_id=FeatureId.NETWORK_MANAGEMENT,
            action_id=ActionId.GET_AP_ENTRIES,
            response_action_id=ActionId.GET_AP_ENTRIES_RSP,
            request_proto=proto.RequestGetApEntries,
            response_proto=proto.ResponseGetApEntries,
            additional_parsers=[BleParserBuilders.MapKey("scan_entry_flags", lambda x: Params.ScanEntry(x))],
        )

        class RequestWifiConnect(BleProtoCommand):
            def __call__(self, ssid: str, password: str) -> GoProResp:
                """Request the camera to connect to a WiFi network.

                Updates will be sent as :py:attr:`open_gopro.constants.ActionId.NOTIF_PROVIS_STATE`

                Args:
                    ssid (str): SSID to connect to
                    password (str): password of WiFi network

                Returns:
                    GoProResp: Command status of request
                """
                return super().__call__(ssid=ssid, password=password)

        #: Sphinx docstring redirect
        self.request_wifi_connect = RequestWifiConnect(
            communicator,
            uuid=GoProUUIDs.CM_NET_MGMT_COMM,
            feature_id=FeatureId.NETWORK_MANAGEMENT,
            action_id=ActionId.REQUEST_WIFI_CONNECT,
            response_action_id=ActionId.REQUEST_WIFI_CONNECT_RSP,
            request_proto=proto.RequestConnectNew,
            response_proto=proto.ResponseConnectNew,
            additional_matching_ids={ActionId.REQUEST_WIFI_CONNECT_RSP},
        )

        class SetLivestreamMode(BleProtoCommand):
            def __call__(
                self,
                url: str,
                window_size: Params.WindowSize,
                cert: bytes,
                minimum_bitrate: int,
                maximum_bitrate: int,
                starting_bitrate: int,
                lens: Params.LensType,
            ) -> GoProResp:
                """Initiate livestream to any site that accepts an RTMP URL and simultaneously encode to camera.

                Args:
                    url (str): url used to stream. Set to empty string to invalidate/cancel stream
                    window_size (open_gopro.api.params.WindowSize): Streaming video resolution
                    cert (bytes): Certificate from a trusted root for streaming services that use encryption
                    minimum_bitrate (int): Desired minimum streaming bitrate (>= 800)
                    maximum_bitrate (int): Desired maximum streaming bitrate (<= 8000)
                    starting_bitrate (int): Initial streaming bitrate (honored if 800 <= value <= 8000)
                    lens (open_gopro.api.params.LensType): Streaming Field of View

                Returns:
                    GoProResp: command status of request
                """
                return super().__call__(
                    url=url,
                    encode=True,
                    window_size=window_size,
                    cert=cert,
                    minimum_bitrate=minimum_bitrate,
                    maximum_bitrate=maximum_bitrate,
                    starting_bitrate=starting_bitrate,
                    lens=lens,
                )

        #: Sphinx docstring redirect
        self.set_livestream_mode = SetLivestreamMode(
            communicator,
            uuid=GoProUUIDs.CQ_COMMAND,
            feature_id=FeatureId.COMMAND,
            action_id=ActionId.SET_LIVESTREAM_MODE,
            response_action_id=ActionId.SET_LIVESTREAM_MODE_RSP,
            request_proto=proto.RequestSetLiveStreamMode,
            response_proto=proto.ResponseGeneric,
        )

        class LivestreamStatus(BleProtoCommand):
            def __call__(
                self,
                register: Optional[list[Params.RegisterLiveStream]] = None,
                unregister: Optional[list[Params.RegisterLiveStream]] = None,
            ) -> GoProResp:
                """Register / unregister to receive asynchronous livestream statuses

                Args:
                    register (Optional[list[open_gopro.api.params.RegisterLiveStream]]): Statuses to register
                        for. Defaults to None (don't register for any).
                    unregister (Optional[list[open_gopro.api.params.RegisterLiveStream]]): Statues to
                        unregister for. Defaults to None (don't unregister for any).

                Returns:
                    GoProResp: current livestream status
                """
                return super().__call__(
                    register_live_stream_status=register or [], unregister_live_stream_status=unregister or []
                )

        #: Sphinx docstring redirect
        self.register_livestream_status = LivestreamStatus(
            communicator,
            uuid=GoProUUIDs.CQ_QUERY,
            feature_id=FeatureId.QUERY,
            action_id=ActionId.GET_LIVESTREAM_STATUS,
            response_action_id=ActionId.LIVESTREAM_STATUS_RSP,
            request_proto=proto.RequestGetLiveStreamStatus,
            response_proto=proto.NotifyLiveStreamStatus,
            additional_matching_ids={ActionId.LIVESTREAM_STATUS_NOTIF},
        )

        class ReleaseNetwork(BleProtoCommand):
            def __call__(self) -> GoProResp:
                """Disconnect the camera Wifi network in STA mode so that it returns to AP mode.

                Returns:
                    GoProResp: status of release request
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.release_network = ReleaseNetwork(
            communicator,
            uuid=GoProUUIDs.CQ_COMMAND,
            feature_id=FeatureId.COMMAND,
            action_id=ActionId.RELEASE_NETWORK,
            response_action_id=ActionId.RELEASE_NETWORK_RSP,
            request_proto=proto.RequestReleaseNetwork,
            response_proto=proto.ResponseGeneric,
        )

        super().__init__(communicator)


class BleSettings(Commands[BleSetting, SettingId]):
    # pylint: disable=missing-class-docstring, unused-argument
    """The collection of all BLE Settings.

    To be used by a GoProBle delegate to build setting commands.

    Args:
        communicator (GoProBle): Adapter to read / write settings
        params: (type[Params]): the set of parameters to use to build the settings
    """

    def __init__(self, communicator: GoProBle):

        self.resolution = BleSetting[Params.Resolution](communicator, SettingId.RESOLUTION, Params.Resolution)
        """Resolution. Set with :py:class:`open_gopro.api.params.Resolution`"""

        self.fps = BleSetting[Params.FPS](communicator, SettingId.FPS, Params.FPS)
        """Frames per second. Set with :py:class:`open_gopro.api.params.FPS`"""

        self.auto_off = BleSetting[Params.AutoOff](communicator, SettingId.AUTO_OFF, Params.AutoOff)
        """Set the auto off time. Set with :py:class:`open_gopro.api.params.AutoOff`"""

        self.video_field_of_view = BleSetting[Params.VideoFOV](
            communicator, SettingId.VIDEO_FOV, Params.VideoFOV
        )
        """Video FOV. Set with :py:class:`open_gopro.api.params.VideoFOV`"""

        self.photo_field_of_view = BleSetting[Params.PhotoFOV](
            communicator, SettingId.PHOTO_FOV, Params.PhotoFOV
        )
        """Photo FOV. Set with :py:class:`open_gopro.api.params.PhotoFOV`"""

        self.multi_shot_field_of_view = BleSetting[Params.MultishotFOV](
            communicator, SettingId.MULTI_SHOT_FOV, Params.MultishotFOV
        )
        """Multi-shot FOV. Set with :py:class:`open_gopro.api.params.MultishotFOV`"""

        self.led = BleSetting[Params.LED](communicator, SettingId.LED, Params.LED)
        """Set the LED options (or also send the BLE keep alive signal). Set with :py:class:`open_gopro.api.params.LED`"""

        self.max_lens_mode = BleSetting[Params.MaxLensMode](
            communicator, SettingId.MAX_LENS_MOD, Params.MaxLensMode
        )
        """Enable / disable max lens mod. Set with :py:class:`open_gopro.api.params.MaxLensMode`"""

        self.hypersmooth = BleSetting[Params.HypersmoothMode](
            communicator, SettingId.HYPERSMOOTH, Params.HypersmoothMode
        )
        """Set / disable hypersmooth. Set with :py:class:`open_gopro.api.params.HypersmoothMode`"""

        self.video_performance_mode = BleSetting[Params.PerformanceMode](
            communicator,
            SettingId.VIDEO_PERFORMANCE_MODE,
            Params.PerformanceMode,
        )
        """Video Performance Mode. Set with :py:class:`open_gopro.api.params.PerformanceMode`"""

        self.media_format = BleSetting[Params.MediaFormat](
            communicator, SettingId.MEDIA_FORMAT, Params.MediaFormat
        )
        """Set the media format. Set with :py:class:`open_gopro.api.params.MediaFormat`"""

        self.anti_flicker = BleSetting[Params.AntiFlicker](
            communicator,
            SettingId.ANTI_FLICKER,
            Params.AntiFlicker,
        )
        """Anti Flicker frequency. Set with :py:class:`open_gopro.api.params.AntiFlicker`"""

        self.camera_ux_mode = BleSetting[Params.CameraUxMode](
            communicator,
            SettingId.CAMERA_UX_MODE,
            Params.CameraUxMode,
        )
        """Camera controls configuration. Set with :py:class:`open_gopro.api.params.CameraUxMode`"""

        self.video_easy_mode = BleSetting[Params.Speed](
            communicator,
            SettingId.VIDEO_EASY_MODE,
            Params.Speed,
        )
        """Video easy mode speed. Set with :py:class:`open_gopro.api.params.Speed`"""

        self.photo_easy_mode = BleSetting[Params.PhotoEasyMode](
            communicator,
            SettingId.PHOTO_EASY_MODE,
            Params.PhotoEasyMode,
        )
        """Night Photo easy mode. Set with :py:class:`open_gopro.api.params.PhotoEasyMode`"""

        self.wifi_band = BleSetting[Params.WifiBand](
            communicator,
            SettingId.WIFI_BAND,
            Params.WifiBand,
        )
        """Current WiFi band being used. Set with :py:class:`open_gopro.api.params.WifiBand`"""

        self.star_trail_length = BleSetting[Params.StarTrailLength](
            communicator,
            SettingId.STAR_TRAIL_LENGTH,
            Params.StarTrailLength,
        )
        """Multi shot star trail length. Set with :py:class:`open_gopro.api.params.StarTrailLength`"""

        self.system_video_mode = BleSetting[Params.SystemVideoMode](
            communicator,
            SettingId.SYSTEM_VIDEO_MODE,
            Params.SystemVideoMode,
        )
        """System video mode. Set with :py:class:`open_gopro.api.params.SystemVideoMode`"""

        self.video_horizon_leveling = BleSetting[Params.HorizonLeveling](
            communicator,
            SettingId.VIDEO_HORIZON_LEVELING,
            Params.HorizonLeveling,
        )
        """Lock / unlock horizon leveling for video. Set with :py:class:`open_gopro.api.params.HorizonLeveling`"""

        self.photo_horizon_leveling = BleSetting[Params.HorizonLeveling](
            communicator,
            SettingId.PHOTO_HORIZON_LEVELING,
            Params.HorizonLeveling,
        )
        """Lock / unlock horizon leveling for photo. Set with :py:class:`open_gopro.api.params.HorizonLeveling`"""

        super().__init__(communicator)


class BleAsyncResponses:
    """These are responses whose ID's are not associated with any commands"""

    generic_response: Final = Struct("unparsed" / GreedyBytes)

    responses = [
        BleAsyncResponse(
            FeatureId.NETWORK_MANAGEMENT, ActionId.NOTIF_PROVIS_STATE, proto.NotifProvisioningState
        ),
        BleAsyncResponse(FeatureId.NETWORK_MANAGEMENT, ActionId.NOTIF_START_SCAN, proto.NotifStartScanning),
        BleAsyncResponse(FeatureId.QUERY, ActionId.INTERNAL_FF, generic_response),
    ]

    @classmethod
    def add_parsers(cls) -> None:
        """Add all of the defined asynchronous responses to the global parser map"""
        for response in cls.responses:
            GoProResp._add_global_parser(response.action_id, response.parser)
            GoProResp._add_feature_action_id_mapping(response.feature_id, response.action_id)


class BleStatuses(Commands[BleStatus, StatusId]):
    """All of the BLE Statuses.

    To be used by a GoProBle delegate to build status commands.

    Args:
        communicator (GoProBle): Adapter to read / write settings
        params: (type[Params]): the set of parameters to use to build the statuses
    """

    def __init__(self, communicator: GoProBle) -> None:
        self.batt_present: BleStatus = BleStatus(communicator, StatusId.BATT_PRESENT, Flag)
        """Is the system's internal battery present?"""

        self.batt_level: BleStatus = BleStatus(communicator, StatusId.BATT_LEVEL, Int8ub)
        """Rough approximation of internal battery level in bars."""

        self.ext_batt_present: BleStatus = BleStatus(communicator, StatusId.EXT_BATT_PRESENT, Flag)
        """Is an external battery connected?"""

        self.ext_batt_level: BleStatus = BleStatus(communicator, StatusId.EXT_BATT_LEVEL, Int8ub)
        """External battery power level in percent."""

        self.system_hot: BleStatus = BleStatus(communicator, StatusId.SYSTEM_HOT, Flag)
        """Is the system currently overheating?"""

        self.system_busy: BleStatus = BleStatus(communicator, StatusId.SYSTEM_BUSY, Flag)
        """Is the camera busy?"""

        self.quick_capture: BleStatus = BleStatus(communicator, StatusId.QUICK_CAPTURE, Flag)
        """Is quick capture feature enabled?"""

        self.encoding_active: BleStatus = BleStatus(communicator, StatusId.ENCODING, Flag)
        """Is the camera currently encoding (i.e. capturing photo / video)?"""

        self.lcd_lock_active: BleStatus = BleStatus(communicator, StatusId.LCD_LOCK_ACTIVE, Flag)
        """Is the LCD lock currently active?"""

        self.video_progress: BleStatus = BleStatus(communicator, StatusId.VIDEO_PROGRESS, Int32ub)
        """When encoding video, this is the duration (seconds) of the video so far; 0 otherwise."""

        self.wireless_enabled: BleStatus = BleStatus(communicator, StatusId.WIRELESS_ENABLED, Flag)
        """Are Wireless Connections enabled?"""

        self.pair_state: BleStatus = BleStatus(communicator, StatusId.PAIR_STATE, Params.PairState)
        """What is the pair state?"""

        self.pair_type: BleStatus = BleStatus(communicator, StatusId.PAIR_TYPE, Params.PairType)
        """The last type of pairing that the camera was engaged in."""

        self.pair_time: BleStatus = BleStatus(communicator, StatusId.PAIR_TIME, Int32ub)
        """	Time (milliseconds) since boot of last successful pairing complete action."""

        self.wap_scan_state: BleStatus = BleStatus(communicator, StatusId.WAP_SCAN_STATE, Params.WAPState)
        """State of current scan for Wifi Access Points. Appears to only change for CAH-related scans."""

        self.wap_scan_time: BleStatus = BleStatus(communicator, StatusId.WAP_SCAN_TIME, Int8ub)
        """The time, in milliseconds since boot that the Wifi Access Point scan completed."""

        self.wap_prov_stat: BleStatus = BleStatus(communicator, StatusId.WAP_PROV_STAT, Params.WAPState)
        """Wifi AP provisioning state."""

        self.remote_ctrl_ver: BleStatus = BleStatus(communicator, StatusId.REMOTE_CTRL_VER, Int8ub)
        """What is the remote control version?"""

        self.remote_ctrl_conn: BleStatus = BleStatus(communicator, StatusId.REMOTE_CTRL_CONN, Flag)
        """Is the remote control connected?"""

        self.pair_state2: BleStatus = BleStatus(communicator, StatusId.PAIR_STATE2, Int8ub)
        """Wireless Pairing State."""

        self.wlan_ssid: BleStatus = BleStatus(communicator, StatusId.WLAN_SSID, GreedyString(encoding="utf-8"))
        """Provisioned WIFI AP SSID. On BLE connection, value is big-endian byte-encoded int."""

        self.ap_ssid: BleStatus = BleStatus(communicator, StatusId.AP_SSID, GreedyString(encoding="utf-8"))
        """Camera's WIFI SSID. On BLE connection, value is big-endian byte-encoded int."""

        self.app_count: BleStatus = BleStatus(communicator, StatusId.APP_COUNT, Int8ub)
        """The number of wireless devices connected to the camera."""

        self.preview_enabled: BleStatus = BleStatus(communicator, StatusId.PREVIEW_ENABLED, Flag)
        """Is preview stream enabled?"""

        self.sd_status: BleStatus = BleStatus(communicator, StatusId.SD_STATUS, Params.SDStatus)
        """Primary Storage Status."""

        self.photos_rem: BleStatus = BleStatus(communicator, StatusId.PHOTOS_REM, Int32ub)
        """How many photos can be taken before sdcard is full?"""

        self.video_rem: BleStatus = BleStatus(communicator, StatusId.VIDEO_REM, Int32ub)
        """How many minutes of video can be captured with current settings before sdcard is full?"""

        self.num_group_photo: BleStatus = BleStatus(communicator, StatusId.NUM_GROUP_PHOTO, Int32ub)
        """How many group photos can be taken with current settings before sdcard is full?"""

        self.num_group_video: BleStatus = BleStatus(communicator, StatusId.NUM_GROUP_VIDEO, Int32ub)
        """Total number of group videos on sdcard."""

        self.num_total_photo: BleStatus = BleStatus(communicator, StatusId.NUM_TOTAL_PHOTO, Int32ub)
        """Total number of photos on sdcard."""

        self.num_total_video: BleStatus = BleStatus(communicator, StatusId.NUM_TOTAL_VIDEO, Int32ub)
        """Total number of videos on sdcard."""

        self.deprecated_40: BleStatus = BleStatus(communicator, StatusId.DEPRECATED_40, DeprecatedAdapter())
        """This status is deprecated."""

        self.ota_stat: BleStatus = BleStatus(communicator, StatusId.OTA_STAT, Params.OTAStatus)
        """The current status of Over The Air (OTA) update."""

        self.download_cancel_pend: BleStatus = BleStatus(communicator, StatusId.DOWNLOAD_CANCEL_PEND, Flag)
        """Is download firmware update cancel request pending?"""

        self.mode_group: BleStatus = BleStatus(communicator, StatusId.MODE_GROUP, Int8ub)
        """Current mode group (deprecated in HERO8)."""

        self.locate_active: BleStatus = BleStatus(communicator, StatusId.LOCATE_ACTIVE, Flag)
        """Is locate camera feature active?"""

        self.multi_count_down: BleStatus = BleStatus(communicator, StatusId.MULTI_COUNT_DOWN, Int32ub)
        """The current timelapse interval countdown value (e.g. 5...4...3...2...1...)."""

        self.space_rem: BleStatus = BleStatus(communicator, StatusId.SPACE_REM, Int64ub)
        """Remaining space on the sdcard in Kilobytes."""

        self.streaming_supp: BleStatus = BleStatus(communicator, StatusId.STREAMING_SUPP, Flag)
        """Is streaming supports in current recording/flatmode/secondary-stream?"""

        self.wifi_bars: BleStatus = BleStatus(communicator, StatusId.WIFI_BARS, Int8ub)
        """Wifi signal strength in bars."""

        self.current_time_ms: BleStatus = BleStatus(communicator, StatusId.CURRENT_TIME_MS, Int32ub)
        """System time in milliseconds since system was booted."""

        self.num_hilights: BleStatus = BleStatus(communicator, StatusId.NUM_HILIGHTS, Int8ub)
        """The number of hilights in encoding video (set to 0 when encoding stops)."""

        self.last_hilight: BleStatus = BleStatus(communicator, StatusId.LAST_HILIGHT, Int32ub)
        """Time since boot (msec) of most recent hilight in encoding video (set to 0 when encoding stops)."""

        self.next_poll: BleStatus = BleStatus(communicator, StatusId.NEXT_POLL, Int32ub)
        """The min time between camera status updates (msec). Do not poll for status more often than this."""

        self.analytics_rdy: BleStatus = BleStatus(communicator, StatusId.ANALYTICS_RDY, Params.AnalyticsState)
        """The current state of camera analytics."""

        self.analytics_size: BleStatus = BleStatus(communicator, StatusId.ANALYTICS_SIZE, Int32ub)
        """The size (units??) of the analytics file."""

        self.in_context_menu: BleStatus = BleStatus(communicator, StatusId.IN_CONTEXT_MENU, Flag)
        """Is the camera currently in a contextual menu (e.g. Preferences)?"""

        self.timelapse_rem: BleStatus = BleStatus(communicator, StatusId.TIMELAPSE_REM, Int32ub)
        """How many min of Timelapse video can be captured with current settings before sdcard is full?"""

        self.exposure_type: BleStatus = BleStatus(communicator, StatusId.EXPOSURE_TYPE, Params.ExposureMode)
        """Liveview Exposure Select Mode."""

        self.exposure_x: BleStatus = BleStatus(communicator, StatusId.EXPOSURE_X, Int8ub)
        """Liveview Exposure Select for y-coordinate (percent)."""

        self.exposure_y: BleStatus = BleStatus(communicator, StatusId.EXPOSURE_Y, Int8ub)
        """Liveview Exposure Select for y-coordinate (percent)."""

        self.gps_stat: BleStatus = BleStatus(communicator, StatusId.GPS_STAT, Flag)
        """Does the camera currently have a GPS lock?"""

        self.ap_state: BleStatus = BleStatus(communicator, StatusId.AP_STATE, Flag)
        """Is the Wifi radio enabled?"""

        self.int_batt_per: BleStatus = BleStatus(communicator, StatusId.INT_BATT_PER, Int8ub)
        """Internal battery level (percent)."""

        self.acc_mic_stat: BleStatus = BleStatus(communicator, StatusId.ACC_MIC_STAT, Params.ExposureMode)
        """Microphone Accessory status."""

        self.digital_zoom: BleStatus = BleStatus(communicator, StatusId.DIGITAL_ZOOM, Int8ub)
        """	Digital Zoom level (percent)."""

        self.wireless_band: BleStatus = BleStatus(communicator, StatusId.WIRELESS_BAND, Params.WifiBand)
        """Wireless Band."""

        self.dig_zoom_active: BleStatus = BleStatus(communicator, StatusId.DIG_ZOOM_ACTIVE, Flag)
        """Is Digital Zoom feature available?"""

        self.mobile_video: BleStatus = BleStatus(communicator, StatusId.MOBILE_VIDEO, Flag)
        """Are current video settings mobile friendly? (related to video compression and frame rate)."""

        self.first_time: BleStatus = BleStatus(communicator, StatusId.FIRST_TIME, Flag)
        """Is the camera currently in First Time Use (FTU) UI flow?"""

        self.sec_sd_stat: BleStatus = BleStatus(communicator, StatusId.SEC_SD_STAT, Params.SDStatus)
        """Secondary Storage Status (exclusive to Superbank)."""

        self.band_5ghz_avail: BleStatus = BleStatus(communicator, StatusId.BAND_5GHZ_AVAIL, Flag)
        """Is 5GHz wireless band available?"""

        self.system_ready: BleStatus = BleStatus(communicator, StatusId.SYSTEM_READY, Flag)
        """Is the system ready to accept commands?"""

        self.batt_ok_ota: BleStatus = BleStatus(communicator, StatusId.BATT_OK_OTA, Flag)
        """Is the internal battery charged sufficiently to start Over The Air (OTA) update?"""

        self.video_low_temp: BleStatus = BleStatus(communicator, StatusId.VIDEO_LOW_TEMP, Flag)
        """Is the camera getting too cold to continue recording?"""

        self.orientation: BleStatus = BleStatus(communicator, StatusId.ORIENTATION, Params.Orientation)
        """The rotational orientation of the camera."""

        self.thermal_mit_mode: BleStatus = BleStatus(communicator, StatusId.DEPRECATED_92, DeprecatedAdapter())
        """This status is deprecated."""

        self.zoom_encoding: BleStatus = BleStatus(communicator, StatusId.ZOOM_ENCODING, Flag)
        """Is this camera capable of zooming while encoding (static value based on model, not settings)?"""

        self.flatmode_id: BleStatus = BleStatus(communicator, StatusId.FLATMODE_ID, Params.Flatmode)
        """Current flatmode ID."""

        self.logs_ready: BleStatus = BleStatus(communicator, StatusId.LOGS_READY, Flag)
        """	Are system logs ready to be downloaded?"""

        self.deprecated_92: BleStatus = BleStatus(communicator, StatusId.DEPRECATED_92, DeprecatedAdapter())
        """This status is deprecated."""

        self.video_presets: BleStatus = BleStatus(communicator, StatusId.VIDEO_PRESETS, Int32ub)
        """Current Video Preset (ID)."""

        self.photo_presets: BleStatus = BleStatus(communicator, StatusId.PHOTO_PRESETS, Int32ub)
        """Current Photo Preset (ID)."""

        self.timelapse_presets: BleStatus = BleStatus(communicator, StatusId.TIMELAPSE_PRESETS, Int32ub)
        """	Current Timelapse Preset (ID)."""

        self.presets_group: BleStatus = BleStatus(communicator, StatusId.PRESETS_GROUP, Int32ub)
        """Current Preset Group (ID)."""

        self.active_preset: BleStatus = BleStatus(communicator, StatusId.ACTIVE_PRESET, Int32ub)
        """Currently Preset (ID)."""

        self.preset_modified: BleStatus = BleStatus(communicator, StatusId.PRESET_MODIFIED, Int32ub)
        """Preset Modified Status, which contains an event ID and a preset (group) ID."""

        self.live_burst_rem: BleStatus = BleStatus(communicator, StatusId.LIVE_BURST_REM, Int32ub)
        """How many Live Bursts can be captured before sdcard is full?"""

        self.live_burst_total: BleStatus = BleStatus(communicator, StatusId.LIVE_BURST_TOTAL, Int32ub)
        """Total number of Live Bursts on sdcard."""

        self.capt_delay_active: BleStatus = BleStatus(communicator, StatusId.CAPT_DELAY_ACTIVE, Flag)
        """Is Capture Delay currently active (i.e. counting down)?"""

        self.media_mod_mic_stat: BleStatus = BleStatus(
            communicator, StatusId.MEDIA_MOD_MIC_STAT, Params.MediaModMicStatus
        )
        """Media mod State."""

        self.timewarp_speed_ramp: BleStatus = BleStatus(
            communicator, StatusId.TIMEWARP_SPEED_RAMP, Params.TimeWarpSpeed
        )
        """Time Warp Speed."""

        self.linux_core_active: BleStatus = BleStatus(communicator, StatusId.LINUX_CORE_ACTIVE, Flag)
        """Is the system's Linux core active?"""

        self.camera_lens_type: BleStatus = BleStatus(
            communicator, StatusId.CAMERA_LENS_TYPE, Params.MaxLensMode
        )
        """Camera lens type (reflects changes to setting 162)."""

        self.video_hindsight: BleStatus = BleStatus(communicator, StatusId.VIDEO_HINDSIGHT, Flag)
        """Is Video Hindsight Capture Active?"""

        self.scheduled_preset: BleStatus = BleStatus(communicator, StatusId.SCHEDULED_PRESET, Int32ub)
        """Scheduled Capture Preset ID."""

        self.scheduled_capture: BleStatus = BleStatus(communicator, StatusId.SCHEDULED_CAPTURE, Flag)
        """Is Scheduled Capture set?"""

        self.creating_preset: BleStatus = BleStatus(communicator, StatusId.CREATING_PRESET, Flag)
        """Is the camera in the process of creating a custom preset?"""

        self.media_mod_stat: BleStatus = BleStatus(
            communicator, StatusId.MEDIA_MOD_STAT, Params.MediaModStatus
        )
        """Media Mode Status (bitmasked)."""

        self.turbo_mode: BleStatus = BleStatus(communicator, StatusId.TURBO_MODE, Flag)
        """Is Turbo Transfer active?"""

        self.sd_rating_check_error: BleStatus = BleStatus(communicator, StatusId.SD_RATING_CHECK_ERROR, Flag)
        """Does sdcard meet specified minimum write speed?"""

        self.sd_write_speed_error: BleStatus = BleStatus(communicator, StatusId.SD_WRITE_SPEED_ERROR, Int8ub)
        """Number of sdcard write speed errors since device booted"""

        self.camera_control: BleStatus = BleStatus(
            communicator, StatusId.CAMERA_CONTROL, Params.CameraControlStatus
        )
        """Camera control status ID"""

        self.usb_connected: BleStatus = BleStatus(communicator, StatusId.USB_CONNECTED, Flag)
        """Is the camera connected to a PC via USB?"""

        self.control_allowed_over_usb: BleStatus = BleStatus(communicator, StatusId.CONTROL_OVER_USB, Flag)
        """Is control allowed over USB?"""

        self.total_sd_space_kb: BleStatus = BleStatus(communicator, StatusId.TOTAL_SD_SPACE_KB, Int32ub)
        """Total space taken up on the SD card in kilobytes"""

        super().__init__(communicator)


# pylint: disable = arguments-renamed
class BleParserBuilders:
    """The collection of custom (i.e. not-construct) parsers and / or builders"""

    class DateTime(BytesParser, BytesBuilder):
        """Handle local and non-local datetime parsing / building"""

        def build(
            self, dt: datetime.datetime, tzone: Optional[int] = None, is_dst: Optional[bool] = None
        ) -> bytes:
            """Build bytestream from datetime and optional local arguments

            Args:
                dt (datetime.datetime): date and time
                tzone (Optional[int], optional): timezone (as UTC offset). Defaults to None.
                is_dst (Optional[bool], optional): is daylight savings time?. Defaults to None.

            Returns:
                bytes: _description_
            """
            byte_data = [*Int16ub.build(dt.year), dt.month, dt.day, dt.hour, dt.minute, dt.second]
            if tzone and is_dst:
                byte_data.extend([*Int16sb.build(tzone), *Flag.build(is_dst)])
            return bytes(byte_data)

        def parse(self, data: bytes) -> dict:
            """Parse bytestream into dict of datetime and potential timezone / dst

            Args:
                data (bytes): bytestream to parse

            Returns:
                dict: dict containing datetime
            """
            is_dst_tz = len(data) == 9
            buf = data[1:]
            year = Int16ub.parse(buf[0:2])

            dt = datetime.datetime(year, *[int(x) for x in buf[2:7]])  # type: ignore
            return (
                dict(datetime=dt)
                if is_dst_tz
                else dict(datetime=dt, tzone=Int16sb.parse(buf[7:9]), dst=bool(buf[9]))
            )

    class MapKey(JsonParser):
        """Map all matching keys using the input function"""

        def __init__(self, key: str, func: Callable) -> None:
            self.key = key
            self.func = func
            super().__init__()

        def parse(self, data: dict) -> dict:
            """Use the map key function to transform the dict

            Args:
                data (dict): input json

            Returns:
                dict: json with keys mapped
            """
            map_keys(data, self.key, self.func)
            return data



# controllers.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Wed Aug 17 20:05:18 UTC 2022

"""GUI controllers and associated common functionality"""

from __future__ import annotations
import enum
import queue
import datetime
import asyncio
import logging
from pathlib import Path
from abc import ABC, abstractmethod
from dataclasses import dataclass
import tkinter as tk
from typing import Optional, Callable, Any, Union, no_type_check

import cv2
import PIL.Image
import PIL.ImageTk
import wrapt

from open_gopro.demos.gui import models, views
from open_gopro.util import setup_logging, pretty_print, add_logging_handler

ResponseHandlerType = Callable[[models.Command, models.GoProResp], None]


def create_widget(
    master: tk.Widget, view: type[views.View], controller: Controller, *args: Any, **kwargs: Any
) -> views.View:
    """Create a widget, binding a view to a controller

    Args:
        master (tk.Widget): master to use for widget
        view (type[views.View]): type of view to use
        controller (Controller): controller instance to use
        args (Any): positional arguments
        kwargs (Any): keyword arguments

    Returns:
        views.View: instantiated view (i.e. widget)
    """
    v = view(master, *args, **kwargs)
    controller.bind(v)
    return v


class MissingArgument(Exception):
    """Exception raised when a command is attempted to be sent without all of its arguments"""

    def __init__(self, argument: str) -> None:
        super().__init__(f"Missing required argument for {argument}")


class BadArgumentValue(Exception):
    """Exception raised when an argument fails a validator check"""

    def __init__(self, argument: str) -> None:
        super().__init__(f"Bad value for argument {argument}")


@wrapt.decorator
def background(wrapped: Callable, instance: Controller, args: Any, kwargs: Any) -> None:
    """Perform an action in the background in order to not block the main GUI processing

    Args:
        wrapped (Callable): action to perform
        instance (Controller): controller that owns action
        args (Any): positional arguments
        kwargs (Any): keyword arguments
    """
    instance.loop.create_task(wrapped(*args, **kwargs))


class Controller(ABC):
    """Controller base class

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
    """

    controllers: list[Controller] = []

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        self.loop = loop
        self._logger: Optional[logging.Logger] = None
        Controller.controllers.append(self)

    @property
    def logger(self) -> logging.Logger:
        """Get the controller's logger

        Raises:
            RuntimeError: logger has not yet been configured

        Returns:
            logging.Logger: logger
        """
        if self._logger:
            return self._logger
        raise RuntimeError("Logger not yet configured")

    @logger.setter
    def logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    @no_type_check
    @abstractmethod
    def bind(self, view: views.View) -> None:
        """Bind a view to the controller

        Args:
            view (views.View): view to bind
        """
        raise NotImplementedError

    def as_async(self, action: Callable, *args: Any) -> Any:
        """Create an asynchronous coroutine from a synchronous function / method

        Args:
            action (Callable): function / method
            args (list[any]): arguments to action

        Returns:
            Any: asynchronous coroutine. Should be awaited in an async method.
        """
        return self.loop.run_in_executor(None, action, *args)


class SplashScreen(Controller):
    """Splash screen used to connect a GoPro

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
        model (models.GoProModel): GoPro device
    """

    def __init__(self, loop: asyncio.AbstractEventLoop, model: models.GoProModel) -> None:
        super().__init__(loop)
        self.model = model
        self._are_models_started = False
        self.view: views.SplashScreen

    @property
    def are_models_started(self) -> bool:
        """Are all of the models started?

        Returns:
            bool: yes or no
        """
        return self._are_models_started

    def bind(self, view: views.SplashScreen) -> None:
        """Bind a splash screen view

        Args:
            view (views.SplashScreen): view to bind
        """
        self.view = view
        self.view.create_view(lambda: self.view.after(1, self.open_models))

    async def update(self) -> None:
        """Update the GUI to display changes"""
        self.view.root.update()
        await asyncio.sleep(0)

    @background
    async def open_models(self) -> None:
        """Open models in the background"""
        device = self.view.device_select.get()
        await self.as_async(self.model.start, device)
        self._are_models_started = True


class Menubar(Controller):
    """Menu bar controller

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
        on_quit (Callable): command to call when quit is clicked
    """

    def __init__(self, loop: asyncio.AbstractEventLoop, on_quit: Callable) -> None:
        super().__init__(loop)
        self.quit = on_quit
        self.view: views.Menubar

    def bind(self, view: views.Menubar) -> None:
        """Bind a menubar view

        Args:
            view (views.Menubar): view to bind
        """
        self.view = view
        view.create_view()
        self.view.fileMenu.add_command(label="Exit", command=self.quit)


class StatusBar(Controller):
    """Status bar

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
    """

    class Ble(enum.Enum):
        """BLE status options"""

        IDLE = views.StatusBar.Color.RED
        CONNECTING = views.StatusBar.Color.YELLOW
        CONNECTED = views.StatusBar.Color.GREEN

    class Wifi(enum.Enum):
        """WiFi status options"""

        IDLE = views.StatusBar.Color.RED
        CONNECTING = views.StatusBar.Color.YELLOW
        CONNECTED = views.StatusBar.Color.GREEN

    class Encoding(enum.Enum):
        """Encoding status options"""

        ON = views.StatusBar.Color.RED
        OFF = views.StatusBar.Color.GREEN

    class Ready(enum.Enum):
        """Ready status options"""

        READY = views.StatusBar.Color.GREEN
        BUSY = views.StatusBar.Color.RED

    class Stream(enum.Enum):
        """Stream status options"""

        IDLE = views.StatusBar.Color.RED
        STARTING = views.StatusBar.Color.YELLOW
        READY = views.StatusBar.Color.GREEN

    StatusType = Union[Ble, Wifi, Encoding, Ready, Stream]

    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        super().__init__(loop)
        self.view: views.StatusBar

    def bind(self, view: views.StatusBar) -> None:
        """Bind a StatusBar view

        Args:
            view (views.StatusBar): view to bind
        """
        self.view = view
        view.create_view()

    def update_status(self, status: StatusType) -> None:
        """Update a status. The relevant status icon will be found based on the passed in status

        Args:
            status (StatusType): new status

        Raises:
            ValueError: a status was passed that is not associated with any status icon
        """

        if status in StatusBar.Ble:
            self.view.update_status(self.view.ble_status, status.value, status.name.replace("_", " ").title())
        elif status in StatusBar.Wifi:
            self.view.update_status(self.view.wifi_status, status.value, status.name.replace("_", " ").title())
        elif status in StatusBar.Ready:
            self.view.update_status(
                self.view.ready_status, status.value, status.name.replace("_", " ").title()
            )
        elif status in StatusBar.Encoding:
            self.view.update_status(
                self.view.encoding_status, status.value, status.name.replace("_", " ").title()
            )
        elif status in StatusBar.Stream:
            self.view.update_status(
                self.view.stream_status, status.value, status.name.replace("_", " ").title()
            )
        else:
            raise ValueError(f"No handler for status {status}")

    def handle_updates(self, identifier: enum.Enum, value: Any) -> None:
        """Check to see if an update requires a status change

        Args:
            identifier (enum.Enum): identifier to analyze
            value (Any): value to analyze
        """
        if identifier == models.constants.StatusId.ENCODING:
            self.update_status(StatusBar.Encoding.ON if value else StatusBar.Encoding.OFF)
        elif identifier == models.constants.StatusId.SYSTEM_READY:
            self.update_status(StatusBar.Ready.READY if value else StatusBar.Ready.BUSY)


class Video(Controller):
    """Displays a video source in a tkinter window

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
    """

    # pylint: disable=no-member
    def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
        super().__init__(loop)
        self.image: PIL.Image.Image
        self.photo: PIL.ImageTk.PhotoImage
        self.vid: Any = None
        self.height: int
        self.width: int
        self.status_bar: StatusBar
        self.should_stop = False
        self.view: views.Video

    @property
    def is_open(self) -> bool:
        """Is the video currently open?

        Returns:
            bool: yes if open, no otherwise
        """
        return self.vid is not None and self.vid.isOpened()

    def bind(self, view: views.Video) -> None:
        """Bind a video view

        Args:
            view (views.Video): view to bind
        """
        self.view = view
        self.view.create_view(self.start, self.stop)

    def bind_status_bar(self, controller: StatusBar) -> None:
        """Bind a status bar controller to use for updating statuses

        Args:
            controller (StatusBar): controller to bind
        """
        self.status_bar = controller

    def start(self, source: Optional[str]) -> None:
        """Start playing a video

        If source is not passed, it will be extracted from the input entry

        Args:
            source (Optional[str]): desired video source or None to get from View
        """
        if self.is_open:
            views.popup_message("Error", "Stream is already started. Must stop first.")
            return

        video_source = source or self.view.url_entry.get()
        if video_source in (None, ""):
            views.popup_message("Error", "Missing required stream URL")
            return
        assert video_source
        # See if this is to be interpreted as an int
        self.status_bar.update_status(StatusBar.Stream.STARTING)
        # Open the video source
        self.logger.debug(f"Starting video from: {video_source}")
        self.vid = cv2.VideoCapture(video_source)
        if not self.is_open:
            views.popup_message("Error", f"Unable to open video source {video_source}")
        self.height = self.view.canvas.winfo_height()
        self.width = self.view.canvas.winfo_width()
        self.should_stop = False
        self.status_bar.update_status(StatusBar.Stream.READY)
        self.get_frame()

    def handle_auto_start(self, command: models.Command, response: models.GoProResp) -> None:
        """Auto start live or preview stream

        Args:
            command (models.Command): command that was sent
            response (models.GoProResp): response that was received
        """
        video_source: Optional[str] = None
        if (identifier := str(command._identifier).lower()) == "livestream" and response.is_ok:
            video_source = response["url"]
        elif identifier == "preview stream" and response.is_ok and "start" in (response.endpoint or ""):
            video_source = models.PREVIEW_STREAM_URL

        if video_source:
            if self.is_open:
                self.logger.debug("Stopping current stream due to new stream starting.")
                self.stop()
            self.start(video_source)

    def stop(self) -> None:
        """Stop playing the video"""
        if not self.is_open:
            views.popup_message("Error", "Stream is not started.")
            return
        self.should_stop = True

    def get_frame(self) -> None:
        """Get an individual video frame to display"""
        ret = False
        frame = None
        if self.vid.isOpened():
            ret, frame = self.vid.read()

        if ret and frame is not None:
            frame = cv2.resize(frame, (self.width, self.height))
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            self.image = PIL.Image.fromarray(frame)
            self.photo = PIL.ImageTk.PhotoImage(image=self.image)
            self.view.display_frame(self.photo)

        if self.should_stop:
            self.vid.release()
            self.status_bar.update_status(StatusBar.Stream.IDLE)
        else:
            self.view.after(5, self.get_frame)

    def __del__(self) -> None:
        # Release the video source when the object is destroyed
        if self.vid.isOpened():
            self.vid.release()


class Log(Controller, logging.Handler):
    """Controller to use for displaying logs.

    Is capable of binding multiple views and then selecting per-record which one to use

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
        level (logging._Level, optional): log level to start at. Defaults to logging.INFO.
    """

    def __init__(self, loop: asyncio.AbstractEventLoop, level: int = logging.INFO) -> None:
        Controller.__init__(self, loop)
        logging.Handler.__init__(self, level)
        self.setFormatter(logging.Formatter("%(asctime)s.%(msecs)03d %(message)s", datefmt="%H:%M:%S"))
        self.setLevel(level)
        self.log_q: queue.Queue[logging.LogRecord] = queue.Queue()
        # Configure logging
        try:
            logger = setup_logging(__name__, Path("gui.log"))
        except RuntimeError:
            # logger is already configured, just get it
            logger = logging.getLogger(__name__)
        add_logging_handler(self)
        self.logger = logger
        self.views: list[views.Log] = []

    def bind(self, view: views.Log) -> None:
        """Bind a log view

        Args:
            view (views.Log): view to bind
        """
        view.create_view()
        self.views.append(view)

    def emit(self, record: logging.LogRecord) -> None:
        """Enqueue the log record string for later displaying in GUI thread

        Args:
            record (logging.LogRecord): record to enqueue
        """
        try:
            self.log_q.put_nowait(record)
        except ValueError:
            return

    def log_queued_messages(self, view: views.Log) -> None:
        """Dequeue log records and display them until the queue is empty

        WARNING! This must be called from the main GUI thread

        Args:
            view (views.Log): view to display records in

        Raises:
            RuntimeError: this view is not bound
        """
        if view not in self.views:
            raise RuntimeError("This view has not been bound to the log controller")
        while not self.log_q.empty() and (record := self.log_q.get_nowait()):
            view.create_log_entry(record)


class CommandPallette(Controller):
    """Pallet to hierarchically display commands and allow user to select.

    This control is used to control a command pallette and param form

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
        model (models.GoProModel): model to get commands from
    """

    @dataclass
    class Argument:
        """Get, store, and process a command argument

        Args:
            name (str): name of argument
            getter (Callable): used to get the string value of the argument from a GUI element
            adapter (Callable): used to adapter the argument string value
            validator (Callable): used to validate the adapted value
        """

        name: str
        getter: Callable[[], Any]
        adapter: Callable[[Any], Any]
        validator: Callable[[Any], None]

        def process(self, validate: bool = True) -> Any:
            """Get, adapt, and optionally validate an argument

            Args:
                validate (bool): Should argument value be validated?. Defaults to True.

            Raises:
                MissingArgument: was not able to get the argument value
                BadArgumentValue: validation failed

            Returns:
                Any: argument value
            """
            value = self.getter()
            try:
                value = self.adapter(value)
            except Exception as e:
                raise MissingArgument(self.name) from e
            if validate:
                try:
                    self.validator(value)
                except Exception as e:
                    raise BadArgumentValue(self.name) from e
            return value

    def __init__(self, loop: asyncio.AbstractEventLoop, model: models.GoProModel) -> None:
        Controller.__init__(self, loop)
        self.model = model
        self.active_command: Any = None
        self.active_arguments: list[CommandPallette.Argument] = []
        self.param_form: views.ParamForm
        self.command_pallette: views.CommandPallette
        self.response_handlers: list[ResponseHandlerType] = []

    def _bind_command_pallette(self, view: views.CommandPallette) -> None:
        """Bind command pallet view

        Args:
            view (views.CommandPallette): view to bind
        """
        self.command_pallette = view
        self.command_pallette.create_view(self.model.command_dict)
        self.command_pallette.tv.bind("<Button-1>", self.on_command_selection)

    def _bind_param_form(self, view: views.ParamForm) -> None:
        """Bind param form view

        Args:
            view (views.ParamForm): view to bind
        """
        self.param_form = view
        self.param_form.create_view()

    def bind(self, view: Union[views.CommandPallette, views.ParamForm]) -> None:
        """Bind a view (command pallette or param form)

        Args:
            view (Union[views.CommandPallette, views.ParamForm]): _description_

        Raises:
            TypeError: Invalid view
        """
        if isinstance(view, views.CommandPallette):
            self._bind_command_pallette(view)
        elif isinstance(view, views.ParamForm):
            self._bind_param_form(view)
        else:
            raise TypeError("Only CommandPallette and ParamForms can be bound to CommandPallette Controller")

    def register_response_handler(self, handler: ResponseHandlerType) -> None:
        """Register response handler for command responses

        Args:
            handler (ResponseHandlerType): response handler controller
        """
        self.response_handlers.append(handler)

    def build_param_entries(self, command: models.Command) -> None:
        """Display all params for a given command in the param form

        Args:
            command (CommandType): command to build params for

        Raises:
            Exception: found a parameter with a currently unhandled argument type
        """
        for adapter, validator, arg_type, arg_name in zip(*self.model.get_command_info(command)):
            getter: Callable
            if arg_type is None:
                continue
            if arg_type == bool:
                getter = self.param_form.create_option_menu(arg_name, ["True", "False"])
            elif arg_type in (int, float, str, bytes, bytearray, Path):
                getter = self.param_form.create_entry(arg_name)
            elif issubclass(arg_type, enum.Enum):
                getter = self.param_form.create_option_menu(arg_name, [x.name for x in arg_type])
            elif arg_type == datetime.datetime:
                getters: list[views.GetterType] = []
                getters.append(self.param_form.create_entry("year"))
                getters.append(self.param_form.create_entry("month"))
                getters.append(self.param_form.create_entry("day"))
                getters.append(self.param_form.create_entry("hour"))
                getters.append(self.param_form.create_entry("minute"))
                getters.append(self.param_form.create_entry("second"))
                getter = lambda getters=getters: tuple(int(get()) for get in getters)
                validator = lambda x: isinstance(x, datetime.datetime)
                adapter = lambda x: datetime.datetime(*x)
            else:
                raise Exception("Unexpected argument type")

            self.active_arguments.append(CommandPallette.Argument(arg_name, getter, adapter, validator))

    def on_command_selection(self, event: Any) -> None:
        """Called when a command is selected to create param entries

        Args:
            event (Any): GUI event that prompted command selection
        """
        try:
            item_id = int(self.command_pallette.tv.identify("item", event.x, event.y))
            if item_id >= views.MAX_TREEVIEW_ID:
                return
        except ValueError:  # User clicked outside of command area
            return
        self.param_form.reset_view()
        self.active_command = None
        self.active_arguments.clear()
        command = self.model.commands[item_id]
        self.param_form.create_command(str(command))

        if self.model.is_command(command):
            self.build_param_entries(command)
            self.param_form.create_button("Send Command", self.command_sender_factory(use_args=True))
        else:  # Setting or Status
            if self.model.is_ble(command):
                self.param_form.create_button("Get Value", self.command_sender_factory("get_value"))
                self.param_form.create_button(
                    "Register for Value Updates", self.command_sender_factory("register_value_update")
                )
                self.param_form.create_button(
                    "Unregister for Value Updates", self.command_sender_factory("unregister_value_update")
                )
                if self.model.is_setting(command):
                    self.param_form.create_button(
                        "Get Capabilities", self.command_sender_factory("get_capabilities_values")
                    )
                    self.param_form.create_button(
                        "Register for Capability Updates",
                        self.command_sender_factory("register_capability_update"),
                    )

                    self.param_form.create_button(
                        "Unregister for Capability Updates",
                        self.command_sender_factory("unregister_capability_update"),
                    )
            if not self.model.is_status(command):
                self.param_form.create_button("Set Value", self.command_sender_factory("set", use_args=True))
                self.build_param_entries(command)

        self.active_command = command

    def command_sender_factory(self, attribute: Optional[str] = None, use_args: bool = False) -> Callable:
        """Build a method to be called when a command is requested to be sent

        Args:
            attribute (Optional[str], optional): Method of active command to use. Defaults to None (use active
                command directly).
            use_args (bool): Should the active args be passed to the command?. Defaults to False.

        Returns:
            Callable: method to use
        """

        @background
        async def on_command_send(self: CommandPallette) -> Optional[models.GoProResp]:
            method = self.active_command if attribute is None else getattr(self.active_command, attribute)
            values = [argument.process() for argument in self.active_arguments] if use_args else []
            try:
                response = await self.as_async(method, *values)
            except Exception as e:  # pylint: disable=broad-except
                views.popup_message("Error", str(e))
                return None

            for handler in self.response_handlers:
                handler(self.active_command, response)
            return response

        return on_command_send.__get__(self, None)


class StatusTab(Controller):
    """Status tab controller to display updates in various status elements.

    New updates will be highlighted

    Args:
        loop (asyncio.AbstractEventLoop): already running asyncio loop to use for controllers
        model (models.GoProModel): model to build updates from
        poll_period (int, optional): how often to refresh updates (in ms). Defaults to 200.
    """

    def __init__(
        self, loop: asyncio.AbstractEventLoop, model: models.GoProModel, poll_period: int = 200
    ) -> None:
        super().__init__(loop)
        self.model = model
        self.period = poll_period
        self.statusbar: Optional[StatusBar] = None
        self.view: views.StatusTab

    def bind(self, view: views.StatusTab) -> None:
        """Bind a status tab view

        Args:
            view (views.StatusTab): view to bind
        """
        self.view = view
        view.create_view()

    def bind_statusbar(self, statusbar: StatusBar) -> None:
        """Bind a status bar

        Args:
            statusbar (StatusBar): view to bind
        """
        self.statusbar = statusbar

    def _display_updates(self, response: Optional[models.GoProResp] = None) -> None:
        """Display any updates take from the response

        Args:
            response (Optional[models.GoProResp], optional): response to get updates from. Defaults to None.
                If none, model will check for asynchronous updates
        """

        cleared = False
        for identifier, value, update_type in self.model.updates(response):
            # Clear recents
            if not cleared:
                self.view.clear_recent_tags()
                cleared = True

            # A list must be a capability update
            if update_type is models.GoProModel.Update.CAPABILITY:
                self.view.update_capability(identifier, pretty_print(value))
                continue
            if update_type is models.GoProModel.Update.PROTOBUF:
                continue
            if update_type is models.GoProModel.Update.SETTING:
                self.view.update_setting(identifier, pretty_print(value))
            elif update_type is models.GoProModel.Update.STATUS:
                self.view.update_status(identifier, pretty_print(value))

            if self.statusbar:
                self.statusbar.handle_updates(identifier, value)

    def display_async_updates(self) -> None:
        """Display any updates that were received asynchronously (i.e. not command responses)"""
        self._display_updates()

    def display_response_updates(self, _: models.Command, response: Optional[models.GoProResp]) -> None:
        """Display updates from the command response

        Args:
            response (Optional[models.GoProResp]): response to get updates from
        """
        self._display_updates(response)

    def poll(self) -> None:
        """Display updates every poll period ms"""
        self.display_async_updates()
        self.view.after(self.period, self.poll)



# wifi_commands.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http:/gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:52 UTC 2021

"""WiFi API for Open GoPro version 1.0"""

from __future__ import annotations
import logging
import datetime
from pathlib import Path
from typing import Any, Optional

from open_gopro.interface import GoProWifi, WifiCommand, Commands
from open_gopro.constants import SettingId, StatusId, CmdId
from open_gopro.responses import GoProResp, JsonParser
from open_gopro.api.builders import WifiGetJsonCommand, WifiGetBinary, WifiSetting
from . import params as Params

logger = logging.getLogger(__name__)


class WifiCommands(Commands[WifiCommand, CmdId]):
    """All of the Wifi commands.

    To be used as a delegate for a GoProWifi to build commands

    All of these return a GoProResp
    """

    # pylint: disable = missing-class-docstring, arguments-differ, useless-super-delegation
    def __init__(self, communicator: GoProWifi):
        """Constructor

        Args:
            communicator (GoProWifi):  Adapter to read / write commands
        """
        ######################################################################################################
        #                          WIFI GET JSON COMMANDS
        ######################################################################################################

        class SetZoom(WifiGetJsonCommand):
            def __call__(self, percent: int) -> GoProResp:
                """Set digital zoom in percent.

                Args:
                    percent (int): Desired zoom as a percentage

                Returns:
                    GoProResp: command status
                """
                return super().__call__(percent=percent)

        #: Sphinx docstring redirect
        self.set_digital_zoom = SetZoom(
            communicator,
            endpoint="gopro/camera/digital_zoom",
            arguments=["percent"],
        )

        class GetCameraState(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get all camera statuses and settings

                Returns:
                    GoProResp: status and settings as JSON
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_camera_state = GetCameraState(
            communicator,
            endpoint="gopro/camera/state",
            parser=WifiParsers.CameraStateParser(),
        )

        class KeepAlive(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Send the keep alive signal to maintain the connection.

                Returns:
                    GoProResp: command status
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.set_keep_alive = KeepAlive(
            communicator,
            endpoint="gopro/camera/keep_alive",
        )

        class GetMediaInfo(WifiGetJsonCommand):
            def __call__(self, file: str) -> GoProResp:
                """Get media info for a file.

                Args:
                    file (str): Media file to get info for

                Returns:
                    GoProResp: Media info as JSON
                """
                return super().__call__(path=f"100GOPRO/{file}")

        #: Sphinx docstring redirect
        self.get_media_info = GetMediaInfo(
            communicator,
            endpoint="gopro/media/info",
            arguments=["path"],
        )

        class GetMediaList(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get a list of media on the camera.

                Returns:
                    GoProResp: Media list JSON structure
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_media_list = GetMediaList(
            communicator,
            endpoint="gopro/media/list",
            parser=WifiParsers.MediaListParser(),
        )

        class SetTurboMode(WifiGetJsonCommand):
            def __call__(self, mode: Params.Toggle) -> GoProResp:
                """Enable or disable Turbo transfer mode.

                Args:
                    mode (open_gopro.api.params.Toggle): enable / disable turbo mode

                Returns:
                    GoProResp: _description_
                """
                return super().__call__(p=mode)

        #: Sphinx docstring redirect
        self.set_turbo_mode = SetTurboMode(
            communicator,
            endpoint="gopro/media/turbo_transfer",
            arguments=["p"],
        )

        class GetOgpVersion(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get Open GoPro API version

                Returns:
                    GoProResp: Open GoPro Version
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_open_gopro_api_version = GetOgpVersion(
            communicator,
            endpoint="gopro/version",
        )

        class GetPresetStatus(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get status of current presets

                Returns:
                    GoProResp: JSON describing currently available presets and preset groups
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_preset_status = GetPresetStatus(
            communicator,
            endpoint="gopro/camera/presets/get",
        )

        class SetPreset(WifiGetJsonCommand):
            def __call__(self, preset: int) -> GoProResp:
                """Set camera to a given preset

                The preset ID can be found from :py:class:`open_gopro.api.wifi_commands.WifiCommands.get_preset_status`

                Args:
                    preset (int): preset to load

                Returns:
                    GoProResp: command status
                """
                return super().__call__(id=preset)

        #: Sphinx docstring redirect
        self.set_preset = SetPreset(
            communicator,
            endpoint="gopro/camera/presets/load",
            arguments=["id"],
        )

        class SetPresetGroup(WifiGetJsonCommand):
            def __call__(self, group: Params.PresetGroup) -> GoProResp:
                """Set the active preset group.

                The most recently used Preset in this group will be set.

                Args:
                    group (open_gopro.api.params.PresetGroup): desired Preset Group

                Returns:
                    GoProResp: command status
                """
                return super().__call__(id=group)

        #: Sphinx docstring redirect
        self.set_preset_group = SetPresetGroup(
            communicator,
            endpoint="gopro/camera/presets/set_group",
            arguments=["id"],
        )

        class SetPreviewStream(WifiGetJsonCommand):
            def __call__(self, mode: Params.Toggle) -> GoProResp:
                """Start or stop the preview stream

                Args:
                    mode (open_gopro.api.params.Toggle): enable to start or disable to stop

                Returns:
                    GoProResp: command status
                """
                return super().__call__(mode="start" if mode is Params.Toggle.ENABLE else "stop")

        #: Sphinx docstring redirect
        self.set_preview_stream = SetPreviewStream(
            communicator,
            endpoint="gopro/camera/stream",
            components=["mode"],
            identifier="Preview Stream",
        )

        class SetThirdPartyInfo(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Flag as third party app

                Returns:
                    GoProResp: command status
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.set_third_party_client_info = SetThirdPartyInfo(
            communicator,
            endpoint="gopro/camera/analytics/set_client_info",
        )

        class SetShutter(WifiGetJsonCommand):
            def __call__(self, shutter: Params.Toggle) -> GoProResp:
                """Set the shutter on or off

                Args:
                    shutter (open_gopro.api.params.Toggle): on or off (i.e. start or stop encoding)

                Returns:
                    GoProResp: command status
                """
                return super().__call__(mode="start" if shutter is Params.Toggle.ENABLE else "stop")

        #: Sphinx docstring redirect
        self.set_shutter_on = SetShutter(
            communicator,
            endpoint="gopro/camera/shutter",
            components=["mode"],
        )

        class SetCameraControl(WifiGetJsonCommand):
            def __call__(self, mode: Params.CameraControl) -> GoProResp:
                """Configure global behaviors by setting camera control (to i.e. Idle, External)

                Args:
                    mode (open_gopro.api.params.CameraControl): desired camera control value

                Returns:
                    GoProResp: command status
                """
                return super().__call__(p=mode)

        #: Sphinx docstring redirect
        self.set_camera_control = SetCameraControl(
            communicator,
            endpoint="gopro/camera/control/set_ui_controller",
            arguments=["p"],
        )

        class SetDateTime(WifiGetJsonCommand):
            def __call__(
                self,
                date_time: datetime.datetime,
                tz_offset: int = 0,
                is_dst: bool = False,
            ) -> GoProResp:
                """Update the date and time of the camera

                Args:
                    date_time (datetime.datetime): date and time
                    tz_offset (int): timezone (as UTC offset). Defaults to 0.
                    is_dst (bool): is daylight savings time?. Defaults to False.

                Returns:
                    GoProResp: command status
                """
                return super().__call__(
                    date=f"{date_time.year}_{date_time.month}_{date_time.day}",
                    time=f"{date_time.hour}_{date_time.minute}_{date_time.second}",
                    tzone=tz_offset,
                    dst=int(is_dst),
                )

        #: Sphinx docstring redirect
        self.set_date_time = SetDateTime(
            communicator,
            endpoint="gopro/camera/set_date_time",
            arguments=["date", "time", "tzone", "dst"],
        )

        class GetDateTime(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get the date and time of the camera (Non timezone / DST aware)

                Returns:
                    GoProResp: current date and time on camera
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_date_time = GetDateTime(
            communicator,
            endpoint="gopro/camera/get_date_time",
        )

        class GetWebcamStatus(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get the status of the webcam endpoint

                Returns:
                    GoProResp: webcam status
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_webcam_status = GetWebcamStatus(
            communicator,
            endpoint="gopro/webcam/status",
        )

        class GetWebcamVersion(WifiGetJsonCommand):
            def __call__(self) -> GoProResp:
                """Get the version of the webcam implementation

                Returns:
                    GoProResp: version
                """
                return super().__call__()

        #: Sphinx docstring redirect
        self.get_webcam_version = GetWebcamVersion(
            communicator,
            endpoint="gopro/webcam/version",
        )

        class AddHilight(WifiGetJsonCommand):
            def __call__(self, file: str, offset: Optional[int] = None) -> GoProResp:
                """Add a hilight to a media file (.mp4)

                Args:
                    file (str):  the media to add the hilight to
                    offset (Optional[int]): offset in ms from start of media

                Returns:
                    GoProResp: command status
                """
                return super().__call__(path=f"100GOPRO/{file}", ms=offset or None)

        #: Sphinx docstring redirect
        self.add_file_hilight = AddHilight(
            communicator,
            endpoint="gopro/media/hilight/file",
            arguments=["path", "ms"],
        )

        class RemoveHilight(WifiGetJsonCommand):
            def __call__(self, file: str, offset: Optional[int] = None) -> GoProResp:
                """Remove a hilight from a media file (.mp4)

                Args:
                    file (str):  the media to remove the hilight from
                    offset (Optional[int]): offset in ms from start of media

                Returns:
                    GoProResp: command status
                """
                return super().__call__(path=f"100GOPRO/{file}", ms=offset)

        #: Sphinx docstring redirect
        self.remove_file_hilight = RemoveHilight(
            communicator,
            endpoint="gopro/media/hilight/remove",
            arguments=["path", "ms"],
        )

        ######################################################################################################
        #                          WIFI GET BINARY COMMANDS
        ######################################################################################################

        class GetGpmfData(WifiGetBinary):
            def __call__(self, camera_file: str, local_file: Optional[Path] = None) -> Path:
                """Get GPMF data for a file.

                If local_file is none, the output location will be the same name as the camera_file.

                Args:
                    camera_file (str): filename on camera to operate on
                    local_file (Optional[Path]): Location on computer to write output. Defaults to None.

                Returns:
                    Path: Path to local_file that output was written to
                """
                return super().__call__(camera_file, local_file)

        #: Sphinx docstring redirect
        self.get_gpmf_data = GetGpmfData(communicator, endpoint="gopro/media/gpmf?path=100GOPRO")

        class GetScreennail(WifiGetBinary):
            def __call__(self, camera_file: str, local_file: Optional[Path] = None) -> Path:
                """Get screennail for a file.

                If local_file is none, the output location will be the same name as the camera_file.

                Args:
                    camera_file (str): filename on camera to operate on
                    local_file (Optional[Path]): Location on computer to write output. Defaults to None.

                Returns:
                    Path: Path to local_file that output was written to
                """
                return super().__call__(camera_file, local_file)

        #: Sphinx docstring redirect
        self.get_screennail = GetScreennail(communicator, endpoint="gopro/media/screennail?path=100GOPRO")

        class GetThumbnail(WifiGetBinary):
            def __call__(self, camera_file: str, local_file: Optional[Path] = None) -> Path:
                """Get thumbnail for a file.

                If local_file is none, the output location will be the same name as the camera_file.

                Args:
                    camera_file (str): filename on camera to operate on
                    local_file (Optional[Path]): Location on computer to write output. Defaults to None.

                Returns:
                    Path: Path to local_file that output was written to
                """
                return super().__call__(camera_file, local_file)

        #: Sphinx docstring redirect
        self.get_thumbnail = GetThumbnail(communicator, endpoint="gopro/media/thumbnail?path=100GOPRO")

        class GetTelemetry(WifiGetBinary):
            def __call__(self, camera_file: str, local_file: Optional[Path] = None) -> Path:
                """Download the telemetry data for a camera file and store in a local file.

                If local_file is none, the output location will be the same name as the camera_file.

                Args:
                    camera_file (str): filename on camera to operate on
                    local_file (Optional[Path]): Location on computer to write output. Defaults to None.

                Returns:
                    Path: Path to local_file that output was written to
                """
                return super().__call__(camera_file, local_file)

        #: Sphinx docstring redirect
        self.get_telemetry = GetTelemetry(communicator, endpoint="gopro/media/telemetry?path=100GOPRO")

        class DownloadFile(WifiGetBinary):
            def __call__(self, camera_file: str, local_file: Optional[Path] = None) -> Path:
                """Download a video from the camera to a local file.

                If local_file is none, the output location will be the same name as the camera_file.

                Args:
                    camera_file (str): filename on camera to operate on
                    local_file (Optional[Path]): Location on computer to write output. Defaults to None.

                Returns:
                    Path: Path to local_file that output was written to
                """
                return super().__call__(camera_file, local_file)

        #: Sphinx docstring redirect
        self.download_file = DownloadFile(
            communicator, endpoint="videos/DCIM/100GOPRO", identifier="Download File"
        )

        super().__init__(communicator)


class WifiSettings(Commands[WifiSetting, SettingId]):
    # pylint: disable=missing-class-docstring, unused-argument
    """The collection of all WiFi Settings

    Args:
        communicator (GoProWifi): Adapter to read / write settings
    """

    def __init__(self, communicator: GoProWifi):
        self.resolution = WifiSetting[Params.Resolution](communicator, SettingId.RESOLUTION)
        """Resolution. Set with :py:class:`open_gopro.api.params.Resolution`"""

        self.fps = WifiSetting[Params.FPS](communicator, SettingId.FPS)
        """Frames per second. Set with :py:class:`open_gopro.api.params.FPS`"""

        self.auto_off = WifiSetting[Params.AutoOff](communicator, SettingId.AUTO_OFF)
        """Set the auto off time. Set with :py:class:`open_gopro.api.params.AutoOff`"""

        self.video_field_of_view = WifiSetting[Params.VideoFOV](communicator, SettingId.VIDEO_FOV)
        """Video FOV. Set with :py:class:`open_gopro.api.params.VideoFOV`"""

        self.photo_field_of_view = WifiSetting[Params.PhotoFOV](communicator, SettingId.PHOTO_FOV)
        """Photo FOV. Set with :py:class:`open_gopro.api.params.PhotoFOV`"""

        self.multi_shot_field_of_view = WifiSetting[Params.MultishotFOV](
            communicator, SettingId.MULTI_SHOT_FOV
        )
        """Multi-shot FOV. Set with :py:class:`open_gopro.api.params.MultishotFOV`"""

        self.max_lens_mode = WifiSetting[Params.MaxLensMode](communicator, SettingId.MAX_LENS_MOD)
        """Enable / disable max lens mod. Set with :py:class:`open_gopro.api.params.MaxLensMode`"""

        self.hypersmooth = WifiSetting[Params.HypersmoothMode](communicator, SettingId.HYPERSMOOTH)
        """Set / disable hypersmooth. Set with :py:class:`open_gopro.api.params.HypersmoothMode`"""

        self.video_performance_mode = WifiSetting[Params.PerformanceMode](
            communicator, SettingId.VIDEO_PERFORMANCE_MODE
        )
        """Video Performance Mode (extended battery, tripod, etc). Set with :py:class:`open_gopro.api.params.PerformanceMode`"""

        self.media_format = WifiSetting[Params.MediaFormat](communicator, SettingId.MEDIA_FORMAT)
        """Set the media format. Set with :py:class:`open_gopro.api.params.MediaFormat`"""

        self.anti_flicker = WifiSetting[Params.AntiFlicker](communicator, SettingId.ANTI_FLICKER)
        """Anti Flicker frequency. Set with :py:class:`open_gopro.api.params.AntiFlicker`"""

        self.camera_ux_mode = WifiSetting[Params.CameraUxMode](communicator, SettingId.CAMERA_UX_MODE)
        """Camera controls configuration. Set with :py:class:`open_gopro.api.params.CameraUxMode`"""

        self.video_easy_mode = WifiSetting[Params.Speed](communicator, SettingId.VIDEO_EASY_MODE)
        """Video easy mode speed. Set with :py:class:`open_gopro.api.params.Speed`"""

        self.photo_easy_mode = WifiSetting[Params.PhotoEasyMode](communicator, SettingId.PHOTO_EASY_MODE)
        """Night Photo easy mode. Set with :py:class:`open_gopro.api.params.PhotoEasyMode`"""

        self.wifi_band = WifiSetting[Params.WifiBand](communicator, SettingId.WIFI_BAND)
        """Current WiFi band being used. Set with :py:class:`open_gopro.api.params.WifiBand`"""

        self.star_trail_length = WifiSetting[Params.StarTrailLength](communicator, SettingId.STAR_TRAIL_LENGTH)
        """Multi shot star trail length. Set with :py:class:`open_gopro.api.params.StarTrailLength`"""

        self.system_video_mode = WifiSetting[Params.SystemVideoMode](communicator, SettingId.SYSTEM_VIDEO_MODE)
        """System video mode. Set with :py:class:`open_gopro.api.params.SystemVideoMode`"""

        self.video_horizon_leveling = WifiSetting[Params.HorizonLeveling](
            communicator, SettingId.VIDEO_HORIZON_LEVELING
        )
        """Lock / unlock horizon leveling for video. Set with :py:class:`open_gopro.api.params.HorizonLeveling`"""

        self.photo_horizon_leveling = WifiSetting[Params.HorizonLeveling](
            communicator, SettingId.PHOTO_HORIZON_LEVELING
        )
        """Lock / unlock horizon leveling for photo. Set with :py:class:`open_gopro.api.params.HorizonLeveling`"""

        super().__init__(communicator)


# pylint: disable = missing-class-docstring
class WifiParsers:
    """The collection of parsers used for additional JSON parsing"""

    class CameraStateParser(JsonParser):
        """Parse integer numbers into Enums"""

        def parse(self, data: dict) -> dict:  # noqa: D102
            parsed: dict[Any, Any] = {}
            # Parse status and settings values into nice human readable things
            for (name, id_map) in [("status", StatusId), ("settings", SettingId)]:
                for k, v in data[name].items():
                    identifier = id_map(int(k))
                    parsed[identifier] = (
                        container(v) if (container := GoProResp._get_query_container(identifier)) else v  # type: ignore
                    )
            return parsed

    class MediaListParser(JsonParser):
        """Extract the list of files from the media list JSON"""

        def parse(self, data: dict) -> dict:  # noqa: D102
            return {"files": data["media"][0]["fs"]}



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file

Defines the structure of protobuf message for network management
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
from . import response_generic_pb2
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions
DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _EnumNetworkOwner:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnumNetworkOwnerEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_EnumNetworkOwner.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DEPRECATED_1: _EnumNetworkOwner.ValueType
    DEPRECATED_2: _EnumNetworkOwner.ValueType
    DEPRECATED_3: _EnumNetworkOwner.ValueType
    DEPRECATED_4: _EnumNetworkOwner.ValueType
    DEPRECATED_5: _EnumNetworkOwner.ValueType
    DEPRECATED_6: _EnumNetworkOwner.ValueType
    DEPRECATED_7: _EnumNetworkOwner.ValueType

class EnumNetworkOwner(_EnumNetworkOwner, metaclass=_EnumNetworkOwnerEnumTypeWrapper): ...

DEPRECATED_1: EnumNetworkOwner.ValueType
DEPRECATED_2: EnumNetworkOwner.ValueType
DEPRECATED_3: EnumNetworkOwner.ValueType
DEPRECATED_4: EnumNetworkOwner.ValueType
DEPRECATED_5: EnumNetworkOwner.ValueType
DEPRECATED_6: EnumNetworkOwner.ValueType
DEPRECATED_7: EnumNetworkOwner.ValueType
global___EnumNetworkOwner = EnumNetworkOwner

class _EnumProvisioning:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnumProvisioningEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_EnumProvisioning.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    PROVISIONING_UNKNOWN: _EnumProvisioning.ValueType
    PROVISIONING_NEVER_STARTED: _EnumProvisioning.ValueType
    PROVISIONING_STARTED: _EnumProvisioning.ValueType
    PROVISIONING_ABORTED_BY_SYSTEM: _EnumProvisioning.ValueType
    PROVISIONING_CANCELLED_BY_USER: _EnumProvisioning.ValueType
    PROVISIONING_SUCCESS_NEW_AP: _EnumProvisioning.ValueType
    PROVISIONING_SUCCESS_OLD_AP: _EnumProvisioning.ValueType
    PROVISIONING_ERROR_FAILED_TO_ASSOCIATE: _EnumProvisioning.ValueType
    PROVISIONING_ERROR_PASSWORD_AUTH: _EnumProvisioning.ValueType
    PROVISIONING_ERROR_EULA_BLOCKING: _EnumProvisioning.ValueType
    PROVISIONING_ERROR_NO_INTERNET: _EnumProvisioning.ValueType
    PROVISIONING_ERROR_UNSUPPORTED_TYPE: _EnumProvisioning.ValueType

class EnumProvisioning(_EnumProvisioning, metaclass=_EnumProvisioningEnumTypeWrapper): ...

PROVISIONING_UNKNOWN: EnumProvisioning.ValueType
PROVISIONING_NEVER_STARTED: EnumProvisioning.ValueType
PROVISIONING_STARTED: EnumProvisioning.ValueType
PROVISIONING_ABORTED_BY_SYSTEM: EnumProvisioning.ValueType
PROVISIONING_CANCELLED_BY_USER: EnumProvisioning.ValueType
PROVISIONING_SUCCESS_NEW_AP: EnumProvisioning.ValueType
PROVISIONING_SUCCESS_OLD_AP: EnumProvisioning.ValueType
PROVISIONING_ERROR_FAILED_TO_ASSOCIATE: EnumProvisioning.ValueType
PROVISIONING_ERROR_PASSWORD_AUTH: EnumProvisioning.ValueType
PROVISIONING_ERROR_EULA_BLOCKING: EnumProvisioning.ValueType
PROVISIONING_ERROR_NO_INTERNET: EnumProvisioning.ValueType
PROVISIONING_ERROR_UNSUPPORTED_TYPE: EnumProvisioning.ValueType
global___EnumProvisioning = EnumProvisioning

class _EnumScanning:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnumScanningEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_EnumScanning.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SCANNING_UNKNOWN: _EnumScanning.ValueType
    SCANNING_NEVER_STARTED: _EnumScanning.ValueType
    SCANNING_STARTED: _EnumScanning.ValueType
    SCANNING_ABORTED_BY_SYSTEM: _EnumScanning.ValueType
    SCANNING_CANCELLED_BY_USER: _EnumScanning.ValueType
    SCANNING_SUCCESS: _EnumScanning.ValueType

class EnumScanning(_EnumScanning, metaclass=_EnumScanningEnumTypeWrapper): ...

SCANNING_UNKNOWN: EnumScanning.ValueType
SCANNING_NEVER_STARTED: EnumScanning.ValueType
SCANNING_STARTED: EnumScanning.ValueType
SCANNING_ABORTED_BY_SYSTEM: EnumScanning.ValueType
SCANNING_CANCELLED_BY_USER: EnumScanning.ValueType
SCANNING_SUCCESS: EnumScanning.ValueType
global___EnumScanning = EnumScanning

class _EnumScanEntryFlags:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _EnumScanEntryFlagsEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_EnumScanEntryFlags.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    SCAN_FLAG_OPEN: _EnumScanEntryFlags.ValueType
    SCAN_FLAG_AUTHENTICATED: _EnumScanEntryFlags.ValueType
    SCAN_FLAG_CONFIGURED: _EnumScanEntryFlags.ValueType
    SCAN_FLAG_BEST_SSID: _EnumScanEntryFlags.ValueType
    SCAN_FLAG_ASSOCIATED: _EnumScanEntryFlags.ValueType
    SCAN_FLAG_UNSUPPORTED_TYPE: _EnumScanEntryFlags.ValueType
    DEPRECATED: _EnumScanEntryFlags.ValueType

class EnumScanEntryFlags(_EnumScanEntryFlags, metaclass=_EnumScanEntryFlagsEnumTypeWrapper): ...

SCAN_FLAG_OPEN: EnumScanEntryFlags.ValueType
SCAN_FLAG_AUTHENTICATED: EnumScanEntryFlags.ValueType
SCAN_FLAG_CONFIGURED: EnumScanEntryFlags.ValueType
SCAN_FLAG_BEST_SSID: EnumScanEntryFlags.ValueType
SCAN_FLAG_ASSOCIATED: EnumScanEntryFlags.ValueType
SCAN_FLAG_UNSUPPORTED_TYPE: EnumScanEntryFlags.ValueType
DEPRECATED: EnumScanEntryFlags.ValueType
global___EnumScanEntryFlags = EnumScanEntryFlags

class NotifProvisioningState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    PROVISIONING_STATE_FIELD_NUMBER: builtins.int
    provisioning_state: global___EnumProvisioning.ValueType
    "Provisioning/connection state"

    def __init__(self, *, provisioning_state: global___EnumProvisioning.ValueType | None = ...) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["provisioning_state", b"provisioning_state"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["provisioning_state", b"provisioning_state"]
    ) -> None: ...

global___NotifProvisioningState = NotifProvisioningState

class NotifStartScanning(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SCANNING_STATE_FIELD_NUMBER: builtins.int
    SCAN_ID_FIELD_NUMBER: builtins.int
    TOTAL_ENTRIES_FIELD_NUMBER: builtins.int
    TOTAL_CONFIGURED_SSID_FIELD_NUMBER: builtins.int
    scanning_state: global___EnumScanning.ValueType
    "Scanning state"
    scan_id: builtins.int
    "ID associated with scan results (included if scan was successful)"
    total_entries: builtins.int
    "Number of APs found during scan (included if scan was successful)"
    total_configured_ssid: builtins.int
    "Total count of camera's provisioned SSIDs"

    def __init__(
        self,
        *,
        scanning_state: global___EnumScanning.ValueType | None = ...,
        scan_id: builtins.int | None = ...,
        total_entries: builtins.int | None = ...,
        total_configured_ssid: builtins.int | None = ...
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "scan_id",
            b"scan_id",
            "scanning_state",
            b"scanning_state",
            "total_configured_ssid",
            b"total_configured_ssid",
            "total_entries",
            b"total_entries",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "scan_id",
            b"scan_id",
            "scanning_state",
            b"scanning_state",
            "total_configured_ssid",
            b"total_configured_ssid",
            "total_entries",
            b"total_entries",
        ],
    ) -> None: ...

global___NotifStartScanning = NotifStartScanning

class RequestConnect(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SSID_FIELD_NUMBER: builtins.int
    OWNER_PURPOSE_FIELD_NUMBER: builtins.int
    ssid: builtins.str
    "AP SSID"
    owner_purpose: global___EnumNetworkOwner.ValueType
    "Deprecated"

    def __init__(
        self,
        *,
        ssid: builtins.str | None = ...,
        owner_purpose: global___EnumNetworkOwner.ValueType | None = ...
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["owner_purpose", b"owner_purpose", "ssid", b"ssid"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["owner_purpose", b"owner_purpose", "ssid", b"ssid"]
    ) -> None: ...

global___RequestConnect = RequestConnect

class RequestConnectNew(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SSID_FIELD_NUMBER: builtins.int
    PASSWORD_FIELD_NUMBER: builtins.int
    STATIC_IP_FIELD_NUMBER: builtins.int
    GATEWAY_FIELD_NUMBER: builtins.int
    SUBNET_FIELD_NUMBER: builtins.int
    DNS_PRIMARY_FIELD_NUMBER: builtins.int
    DNS_SECONDARY_FIELD_NUMBER: builtins.int
    SET_TO_LEAST_PREFERRED_AP_FIELD_NUMBER: builtins.int
    OWNER_PURPOSE_FIELD_NUMBER: builtins.int
    ssid: builtins.str
    "AP SSID"
    password: builtins.str
    "AP password"
    static_ip: builtins.bytes
    "Static IP address"
    gateway: builtins.bytes
    "Gateway IP address"
    subnet: builtins.bytes
    "Subnet mask"
    dns_primary: builtins.bytes
    "Primary DNS"
    dns_secondary: builtins.bytes
    "Secondary DNS"
    set_to_least_preferred_ap: builtins.bool
    "Deprecated"
    owner_purpose: global___EnumNetworkOwner.ValueType
    "Deprecated"

    def __init__(
        self,
        *,
        ssid: builtins.str | None = ...,
        password: builtins.str | None = ...,
        static_ip: builtins.bytes | None = ...,
        gateway: builtins.bytes | None = ...,
        subnet: builtins.bytes | None = ...,
        dns_primary: builtins.bytes | None = ...,
        dns_secondary: builtins.bytes | None = ...,
        set_to_least_preferred_ap: builtins.bool | None = ...,
        owner_purpose: global___EnumNetworkOwner.ValueType | None = ...
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "dns_primary",
            b"dns_primary",
            "dns_secondary",
            b"dns_secondary",
            "gateway",
            b"gateway",
            "owner_purpose",
            b"owner_purpose",
            "password",
            b"password",
            "set_to_least_preferred_ap",
            b"set_to_least_preferred_ap",
            "ssid",
            b"ssid",
            "static_ip",
            b"static_ip",
            "subnet",
            b"subnet",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "dns_primary",
            b"dns_primary",
            "dns_secondary",
            b"dns_secondary",
            "gateway",
            b"gateway",
            "owner_purpose",
            b"owner_purpose",
            "password",
            b"password",
            "set_to_least_preferred_ap",
            b"set_to_least_preferred_ap",
            "ssid",
            b"ssid",
            "static_ip",
            b"static_ip",
            "subnet",
            b"subnet",
        ],
    ) -> None: ...

global___RequestConnectNew = RequestConnectNew

class RequestGetApEntries(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    START_INDEX_FIELD_NUMBER: builtins.int
    MAX_ENTRIES_FIELD_NUMBER: builtins.int
    SCAN_ID_FIELD_NUMBER: builtins.int
    start_index: builtins.int
    "Used for paging. 0 <= start_index < NotifStartScanning.total_entries"
    max_entries: builtins.int
    "Used for paging. Value must be < NotifStartScanning.total_entries"
    scan_id: builtins.int
    "ID corresponding to a set of scan results (i.e."

    def __init__(
        self,
        *,
        start_index: builtins.int | None = ...,
        max_entries: builtins.int | None = ...,
        scan_id: builtins.int | None = ...
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "max_entries", b"max_entries", "scan_id", b"scan_id", "start_index", b"start_index"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "max_entries", b"max_entries", "scan_id", b"scan_id", "start_index", b"start_index"
        ],
    ) -> None: ...

global___RequestGetApEntries = RequestGetApEntries

class RequestReleaseNetwork(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___RequestReleaseNetwork = RequestReleaseNetwork

class RequestStartScan(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(self) -> None: ...

global___RequestStartScan = RequestStartScan

class ResponseConnect(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RESULT_FIELD_NUMBER: builtins.int
    PROVISIONING_STATE_FIELD_NUMBER: builtins.int
    TIMEOUT_SECONDS_FIELD_NUMBER: builtins.int
    result: response_generic_pb2.EnumResultGeneric.ValueType
    "Generic pass/fail/error info"
    provisioning_state: global___EnumProvisioning.ValueType
    "Provisioning/connection state"
    timeout_seconds: builtins.int
    "Network connection timeout (seconds)"

    def __init__(
        self,
        *,
        result: response_generic_pb2.EnumResultGeneric.ValueType | None = ...,
        provisioning_state: global___EnumProvisioning.ValueType | None = ...,
        timeout_seconds: builtins.int | None = ...
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "provisioning_state",
            b"provisioning_state",
            "result",
            b"result",
            "timeout_seconds",
            b"timeout_seconds",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "provisioning_state",
            b"provisioning_state",
            "result",
            b"result",
            "timeout_seconds",
            b"timeout_seconds",
        ],
    ) -> None: ...

global___ResponseConnect = ResponseConnect

class ResponseConnectNew(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RESULT_FIELD_NUMBER: builtins.int
    PROVISIONING_STATE_FIELD_NUMBER: builtins.int
    TIMEOUT_SECONDS_FIELD_NUMBER: builtins.int
    result: response_generic_pb2.EnumResultGeneric.ValueType
    "Status of Connect New request"
    provisioning_state: global___EnumProvisioning.ValueType
    "Current provisioning state of the network"
    timeout_seconds: builtins.int

    def __init__(
        self,
        *,
        result: response_generic_pb2.EnumResultGeneric.ValueType | None = ...,
        provisioning_state: global___EnumProvisioning.ValueType | None = ...,
        timeout_seconds: builtins.int | None = ...
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "provisioning_state",
            b"provisioning_state",
            "result",
            b"result",
            "timeout_seconds",
            b"timeout_seconds",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "provisioning_state",
            b"provisioning_state",
            "result",
            b"result",
            "timeout_seconds",
            b"timeout_seconds",
        ],
    ) -> None: ...

global___ResponseConnectNew = ResponseConnectNew

class ScanEntry(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SSID_FIELD_NUMBER: builtins.int
    SIGNAL_STRENGTH_BARS_FIELD_NUMBER: builtins.int
    SIGNAL_FREQUENCY_MHZ_FIELD_NUMBER: builtins.int
    SCAN_ENTRY_FLAGS_FIELD_NUMBER: builtins.int
    OWNER_PURPOSE_FIELD_NUMBER: builtins.int
    ssid: builtins.str
    "AP SSID"
    signal_strength_bars: builtins.int
    "Signal strength (3 bars: >-70 dBm; 2 bars: >-85 dBm; 1 bar: <=-85"
    signal_frequency_mhz: builtins.int
    "Signal frequency (MHz)"
    scan_entry_flags: builtins.int
    "Bitmasked value from EnumScanEntryFlags"
    owner_purpose: global___EnumNetworkOwner.ValueType
    "Deprecated"

    def __init__(
        self,
        *,
        ssid: builtins.str | None = ...,
        signal_strength_bars: builtins.int | None = ...,
        signal_frequency_mhz: builtins.int | None = ...,
        scan_entry_flags: builtins.int | None = ...,
        owner_purpose: global___EnumNetworkOwner.ValueType | None = ...
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "owner_purpose",
            b"owner_purpose",
            "scan_entry_flags",
            b"scan_entry_flags",
            "signal_frequency_mhz",
            b"signal_frequency_mhz",
            "signal_strength_bars",
            b"signal_strength_bars",
            "ssid",
            b"ssid",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "owner_purpose",
            b"owner_purpose",
            "scan_entry_flags",
            b"scan_entry_flags",
            "signal_frequency_mhz",
            b"signal_frequency_mhz",
            "signal_strength_bars",
            b"signal_strength_bars",
            "ssid",
            b"ssid",
        ],
    ) -> None: ...

global___ScanEntry = ScanEntry

class ResponseGetApEntries(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RESULT_FIELD_NUMBER: builtins.int
    SCAN_ID_FIELD_NUMBER: builtins.int
    ENTRIES_FIELD_NUMBER: builtins.int
    result: response_generic_pb2.EnumResultGeneric.ValueType
    "Generic pass/fail/error info"
    scan_id: builtins.int
    "ID associated with this batch of results"

    @property
    def entries(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ScanEntry]:
        """Array containing details about discovered APs"""
    def __init__(
        self,
        *,
        result: response_generic_pb2.EnumResultGeneric.ValueType | None = ...,
        scan_id: builtins.int | None = ...,
        entries: collections.abc.Iterable[global___ScanEntry] | None = ...
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["result", b"result", "scan_id", b"scan_id"]
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "entries", b"entries", "result", b"result", "scan_id", b"scan_id"
        ],
    ) -> None: ...

global___ResponseGetApEntries = ResponseGetApEntries

class ResponseStartScanning(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    RESULT_FIELD_NUMBER: builtins.int
    SCANNING_STATE_FIELD_NUMBER: builtins.int
    result: response_generic_pb2.EnumResultGeneric.ValueType
    "Generic pass/fail/error info"
    scanning_state: global___EnumScanning.ValueType
    "Scanning state"

    def __init__(
        self,
        *,
        result: response_generic_pb2.EnumResultGeneric.ValueType | None = ...,
        scanning_state: global___EnumScanning.ValueType | None = ...
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["result", b"result", "scanning_state", b"scanning_state"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["result", b"result", "scanning_state", b"scanning_state"]
    ) -> None: ...

global___ResponseStartScanning = ResponseStartScanning



# builders.py/Open GoPro, Version 2.0 (C) Copyright 2021 GoPro, Inc. (http://gopro.com/OpenGoPro).
# This copyright was auto-generated on Tue Sep  7 21:35:53 UTC 2021

"""Common functionality across API versions to build commands, settings, and statuses"""

from __future__ import annotations
import enum
import logging
from pathlib import Path
from abc import abstractmethod
from urllib.parse import urlencode
from collections.abc import Iterable
from dataclasses import dataclass, InitVar, field
from typing import Any, TypeVar, Generic, Union, no_type_check, Optional, Final

import google.protobuf.json_format
from google.protobuf import descriptor
from google.protobuf.message import Message as Protobuf
from google.protobuf.json_format import MessageToDict as ProtobufToDict
import construct

from open_gopro.responses import BytesBuilder, BytesParser, GoProResp, JsonParser, BytesParserBuilder, Parser
from open_gopro.constants import (
    ActionId,
    FeatureId,
    BleUUID,
    CmdId,
    SettingId,
    QueryCmdId,
    StatusId,
    GoProUUIDs,
    enum_factory,
    GoProEnum,
)
from open_gopro.interface import GoProBle, GoProWifi, BleCommand, WifiCommand
from open_gopro.util import Logger, jsonify

logger = logging.getLogger(__name__)

ValueType = TypeVar("ValueType")
IdType = TypeVar("IdType")
ProtobufProducerType = tuple[Union[type[SettingId], type[StatusId]], QueryCmdId]

ProtobufPrinter = google.protobuf.json_format._Printer  # type: ignore # noqa
original_field_to_json = ProtobufPrinter._FieldToJsonObject

QueryParserType = Union[construct.Construct, type[GoProEnum], BytesParserBuilder]
AsyncParserType = Union[construct.Construct, BytesParser[dict], type[Protobuf]]

######################################################## BLE #################################################


def enum_parser_factory(target: type[GoProEnum]) -> BytesParserBuilder:
    """Build an Enum ParserBuilder

    Args:
        target (type[GoProEnum]): enum to use for parsing and building

    Returns:
        BytesParserBuilder: instance of generated class
    """

    class ParserBuilder(BytesParserBuilder[GoProEnum]):
        """Adapt enums to / from a one byte value"""

        container = target

        def parse(self, data: bytes) -> GoProEnum:
            return self.container(data[0])

        def build(self, *args: Any, **kwargs: Any) -> bytes:
            return bytes([int(args[0])])

    return ParserBuilder()


def construct_adapter_factory(target: construct.Construct) -> BytesParserBuilder:
    """Build a construct parser adapter from a construct

    Args:
        target (construct.Construct): construct to use for parsing and building

    Returns:
        BytesParserBuilder: instance of generated class
    """

    class ParserBuilder(BytesParserBuilder):
        """Adapt the construct for our interface"""

        container = target

        def parse(self, data: bytes) -> Any:
            return self.container.parse(data)

        def build(self, *args: Any, **kwargs: Any) -> bytes:
            return self.container.build(*args, **kwargs)

    return ParserBuilder()


def protobuf_parser_factory(proto: type[Protobuf]) -> BytesParser[dict]:
    """Build a BytesParser from a protobuf definition

    Args:
        proto (type[Protobuf]): protobuf definition to build class from

    Returns:
        BytesParser[dict]: instance of generated class
    """

    class ProtobufByteParser(BytesParser[dict]):
        """Parse bytes into a dict using the protobuf"""

        protobuf = proto

        def parse(self, data: bytes) -> dict:
            response: Protobuf = self.protobuf().FromString(bytes(data))

            # Monkey patch the field-to-json function to use our enum translation
            ProtobufPrinter._FieldToJsonObject = (
                lambda self, field, value: enum_factory(field.enum_type)(value)  # pylint: disable=not-callable
                if field.cpp_type == descriptor.FieldDescriptor.CPPTYPE_ENUM
                else original_field_to_json(self, field, value)
            )
            return ProtobufToDict(response, preserving_proto_field_name=True)

    return ProtobufByteParser()


class DeprecatedAdapter(BytesParserBuilder[str]):
    """Used to return "DEPRECATED" when a deprecated setting / status is attempted to be parsed / built"""

    def parse(self, data: bytes) -> str:
        """Return string indicating this ID is deprecated

        Args:
            data (bytes): ignored

        Returns:
            str: "DEPRECATED"
        """
        return "DEPRECATED"

    def build(self, obj: Any) -> bytes:
        """Return empty bytes since this ID is deprecated

        Args:
            obj (Any): ignored

        Returns:
            bytes: empty
        """
        return bytes()


class BleReadCommand(BleCommand[BleUUID]):
    """A BLE command that reads data from a BleUUID

    Args:
        communicator (GoProBle): BLE client to read
        uuid (BleUUID): BleUUID to read from
        response_parser (BytesParser): the parser that will parse the received bytestream into a JSON dict
    """

    def __init__(
        self,
        communicator: GoProBle,
        uuid: BleUUID,
        parser: Optional[Union[construct.Construct, BytesParser[dict]]],
    ) -> None:
        super().__init__(
            communicator,
            uuid=uuid,
            parser=construct_adapter_factory(parser) if isinstance(parser, construct.Construct) else parser,
            identifier=uuid,
        )

    def __call__(self) -> GoProResp:  # noqa: D102
        logger.info(Logger.build_log_tx_str(jsonify(self._as_dict())))
        response = self._communicator._read_characteristic(self._uuid)
        logger.info(Logger.build_log_rx_str(response))
        return response

    def __str__(self) -> str:
        return f"Read {self._uuid.name.lower().replace('_', ' ').title()}"

    def _as_dict(self, *_: Any, **kwargs: Any) -> dict[str, Any]:
        """Return the attributes of the command as a dict

        Args:
            *_ (Any): unused
            **kwargs (Any): additional entries for the dict

        Returns:
            dict[str, Any]: command as dict
        """
        return dict(id="Read " + self._uuid.name, **self._base_dict) | kwargs


class BleWriteCommand(BleCommand[CmdId]):
    """A BLE command that writes to a BleUUID and does not accept any parameters

    Args:
        communicator (GoProBle): BLE client to write
        uuid (BleUUID): BleUUID to write to
        cmd (CmdId): Command ID that is being sent
        param_builder (BytesBuilder, optional): is responsible for building the bytestream to send from the input params
        parser (BytesParser. optional): the parser that will parse the received bytestream into a JSON dict
    """

    def __init__(
        self,
        communicator: GoProBle,
        uuid: BleUUID,
        cmd: CmdId,
        param_builder: Optional[BytesBuilder] = None,
        parser: Optional[Union[construct.Construct, BytesParser[dict]]] = None,
    ) -> None:
        self.param_builder = param_builder
        self.cmd = cmd
        super().__init__(
            communicator,
            uuid=uuid,
            parser=construct_adapter_factory(parser) if isinstance(parser, construct.Construct) else parser,
            identifier=cmd,
        )

    @no_type_check
    def __call__(self, /, **kwargs: Any) -> GoProResp:  # noqa: D102
        logger.info(Logger.build_log_tx_str(jsonify(self._as_dict(**kwargs))))

        data = bytearray([self.cmd.value])
        params = bytearray()
        if self.param_builder:
            params.extend(self.param_builder.build(*kwargs.values()))
        else:
            for arg in kwargs.values():
                params.extend(arg.value if isinstance(arg, enum.Enum) else arg)
        if params:
            data.append(len(params))
            data.extend(params)
        response = self._communicator._send_ble_command(self._uuid, data, self._identifier)
        logger.info(Logger.build_log_rx_str(response))
        return response

    def __str__(self) -> str:
        return self.cmd.name.lower().replace("_", " ").removeprefix("cmdid").title()

    def _as_dict(self, *_: Any, **kwargs: Any) -> dict[str, Any]:
        """Return the attributes of the command as a dict

        Args:
            *_ (Any): unused
            **kwargs (Any): additional entries for the dict

        Returns:
            dict[str, Any]: command as dict
        """
        return dict(id=self.cmd, **self._base_dict) | kwargs


class RegisterUnregisterAll(BleWriteCommand):
    """Base class for register / unregister all commands

    This will loop over all of the elements (i.e. settings / statuses found from the element_set entry of the
    producer tuple parameter) and individually register / unregister (depending on the action parameter) each
    element in the set

    Args:
        producer: (ProtobufProducerType): Tuple of (element_set, query command) where element_set is the GoProEnum
            that this command relates to, i.e. SettingId for settings, StatusId for Statuses
        action: (Action): whether to register or unregister
    """

    class Action(enum.Enum):
        """Enum to differentiate between register actions"""

        REGISTER = enum.auto()
        UNREGISTER = enum.auto()

    def __init__(
        self,
        communicator: GoProBle,
        uuid: BleUUID,
        cmd: CmdId,
        producer: ProtobufProducerType,
        action: Action,
        parser: Optional[BytesParser] = None,
    ) -> None:
        self.action = action
        self.producer = producer
        super().__init__(communicator, uuid=uuid, cmd=cmd, parser=parser)

    def __call__(self, **kwargs: Any) -> GoProResp:  # noqa: D102
        element_set = self.producer[0]
        responded_command = self.producer[1]
        response = super().__call__()
        if response.is_ok:
            for element in element_set:
                (
                    self._communicator._register_listener
                    if self.action is RegisterUnregisterAll.Action.REGISTER
                    else self._communicator._unregister_listener
                )(
                    (responded_command, element)  # type: ignore
                )
        return response


class BleProtoCommand(BleCommand[ActionId]):
    """A BLE command that is sent and received as using the Protobuf protocol

    Args:
        communicator (GoProBle): BLE client to write
        uuid (BleUUID): BleUUID to write to
        feature_id (CmdId): Command ID that is being sent
        action_id (FeatureId): protobuf specific action ID that is being sent
        response_action_id (ActionId):  the action ID that will be in the response
        request_proto (type[Protobuf]): protobuf used to build command bytestream
        response_proto (type[Protobuf]): protobuf used to parse received bytestream
        additional_matching_ids: (Optional[set[ActionId]]): Other action ID's to share
            this parser. This is used, for example, if a notification shares the same ID as the
            synchronous response. Defaults to None.
    """

    def __init__(
        self,
        communicator: GoProBle,
        uuid: BleUUID,
        feature_id: FeatureId,
        action_id: ActionId,
        response_action_id: ActionId,
        request_proto: type[Protobuf],
        response_proto: type[Protobuf],
        additional_matching_ids: Optional[set[Union[ActionId, CmdId]]] = None,
        additional_parsers: Optional[list[JsonParser]] = None,
    ) -> None:
        parser = protobuf_parser_factory(response_proto)
        for p in additional_parsers or []:
            parser += p  # type: ignore
        super().__init__(communicator, uuid=uuid, parser=parser, identifier=action_id)
        self.feature_id = feature_id
        self.action_id = action_id
        self.response_action_id = response_action_id
        self.request_proto = request_proto
        self.response_proto = response_proto
        self.additional_matching_ids: set[Union[ActionId, CmdId]] = additional_matching_ids or set()
        assert self._parser
        for matching_id in [*self.additional_matching_ids, response_action_id]:
            GoProResp._add_global_parser(matching_id, self._parser)
        GoProResp._add_feature_action_id_mapping(self.feature_id, self.response_action_id)

    def build_data(self, **kwargs: Any) -> bytearray:
        """Build the byte data to prepare for command sending

        Args:
            **kwargs (Any): arguments to command to use to build protobuf

        Returns:
            bytearray: built byte data
        """
        proto = self.request_proto()
        for attr_name, arg in kwargs.items():
            value = arg.value if issubclass(type(arg), enum.Enum) else arg
            attr = getattr(proto, attr_name)
            # Protobuf "repeatable" (i.e. iterable) fields can not be set directly and must be appended / extended
            if isinstance(attr, Iterable) and not isinstance(value, (str, bytes)):
                if isinstance(value, Iterable):
                    for element in value:
                        attr.append(element.value if isinstance(element, enum.Enum) else element)  # type: ignore
                else:
                    attr.append(value.value if isinstance(value, enum.Enum) else value)  # type:ignore
            else:
                setattr(proto, attr_name, value)
        # Prepend headers and serialize
        return bytearray([self.feature_id.value, self.action_id.value, *proto.SerializeToString()])

    @abstractmethod
    @no_type_check
    # pylint: disable=missing-return-doc
    def __call__(self, /, **kwargs: Any) -> GoProResp:  # noqa: D102
        # The method that will actually build and send the protobuf command
        logger.info(Logger.build_log_tx_str(jsonify(self._as_dict(**kwargs))))
        data = self.build_data(**kwargs)
        # Allow exception to pass through if protobuf not completely initialized
        response = self._communicator._send_ble_command(self._uuid, data, self.response_action_id)
        logger.info(Logger.build_log_rx_str(response))
        return response

    def __str__(self) -> str:
        return self.action_id.name.lower().replace("_", " ").removeprefix("actionid").title()

    def _as_dict(self, *_: Any, **kwargs: Any) -> dict[str, Any]:
        """Return the attributes of the command as a dict

        Args:
            *_ (Any): unused
            **kwargs (Any): additional entries for the dict

        Returns:
            dict[str, Any]: command as dict
        """
        return dict(id=self.action_id, feature_id=self.feature_id, **self._base_dict) | kwargs


@dataclass
class BleAsyncResponse:
    """A BLE protobuf response that is not associated with any command.

    Args:
        feature_id (FeatureId): Feature ID that response corresponds to
        action_id (ActionId): Action ID that response corresponds to
        parser_type (AsyncParserType): how to parse the response
    """

    feature_id: FeatureId
    action_id: ActionId
    parser_type: InitVar[AsyncParserType]
    parser: Parser = field(init=False)

    def __post_init__(self, parser_type: AsyncParserType) -> None:
        if isinstance(parser_type, construct.Construct):
            self.parser = construct_adapter_factory(parser_type)
        elif isinstance(parser_type, BytesParser):
            self.parser = parser_type
        elif issubclass(parser_type, Protobuf):
            self.parser = protobuf_parser_factory(parser_type)
        else:
            raise TypeError(f"Unexpected {parser_type=}")

    def __str__(self) -> str:
        return self.action_id.name.lower().replace("_", " ").removeprefix("actionid").title()


class BleSetting(BleCommand[SettingId], Generic[ValueType]):
    """An individual camera setting that is interacted with via BLE."""

    SETTER_UUID: Final[BleUUID] = GoProUUIDs.CQ_SETTINGS
    READER_UUID: Final[BleUUID] = GoProUUIDs.CQ_QUERY

    def __init__(self, communicator: GoProBle, identifier: SettingId, parser_builder: QueryParserType) -> None:
        """Constructor

        Args:
            communicator (GoProBle): Adapter to read / write settings data
            identifier (SettingId): ID of setting
            parser_builder (QueryParserType): object to both parse and build setting

        Raises:
            TypeError: Invalid parser_builder type
        """
        if isinstance(parser_builder, construct.Construct):
            parser = construct_adapter_factory(parser_builder)
        elif isinstance(parser_builder, BytesParserBuilder):
            parser = parser_builder
        elif issubclass(parser_builder, GoProEnum):
            parser = enum_parser_factory(parser_builder)
        else:
            raise TypeError(f"Unexpected {parser_builder=}")
        self._identifier = identifier
        self._builder = parser
        BleCommand.__init__(self, communicator, uuid=self.SETTER_UUID, parser=parser, identifier=identifier)

    def __str__(self) -> str:
        return str(self._identifier.name).lower().replace("_", " ").title()

    def _as_dict(  # pylint: disable = arguments-differ
        self, identifier: Union[QueryCmdId, SettingId, str], *_: Any, **kwargs: Any
    ) -> dict[str, Any]:
        """Return the attributes of the command as a dict

        Args:
            identifier (Union[QueryCmdId, SettingId, str]): identifier of the command for this send
            *_ (Any): unused
            **kwargs (Any): additional entries for the dict

        Returns:
            dict[str, Any]: setting as dict
        """
        return dict(id=identifier, **self._base_dict) | kwargs

    def set(self, value: ValueType) -> GoProResp:
        """Set the value of the setting.

        Args:
            value (ValueType): The argument to use to set the setting value.

        Returns:
            GoProResp: Status of set
        """
        logger.info(
            Logger.build_log_tx_str(jsonify(self._as_dict(f"Set {str(self._identifier)}", value=value)))
        )
        # Special case. Can't use _send_query
        data = bytearray([self._identifier.value])
        try:
            param = self._builder.build(value)
            data.extend([len(param), *param])
        except IndexError:
            pass

        response = self._communicator._send_ble_command(self.SETTER_UUID, data, self._identifier)
        logger.info(Logger.build_log_rx_str(response))
        return response

    def _send_query(self, response_id: QueryCmdId) -> GoProResp:
        """Build the byte data and query setting information

        Args:
            response_id (QueryCmdId): expected identifier of response

        Returns:
            GoProResp: query response
        """
        data = self._build_cmd(response_id)
        logger.info(
            Logger.build_log_tx_str(jsonify(self._as_dict(f"{str(response_id)}.{self._identifier.name}")))
        )
        response = self._communicator._send_ble_command(self.READER_UUID, data, response_id)
        logger.info(Logger.build_log_rx_str(response))
        return response

    def get_value(self) -> GoProResp:
        """Get the settings value.

        Returns:
            GoProResp: settings value
        """
        return self._send_query(QueryCmdId.GET_SETTING_VAL)

    def get_name(self) -> GoProResp:
        """Get the settings name.

        Raises:
            NotImplementedError: This isn't implemented on the camera
        """
        raise NotImplementedError("Not implemented on camera!")

    def get_capabilities_values(self) -> GoProResp:
        """Get currently supported settings capabilities values.

        Returns:
            GoProResp: settings capabilities values
        """
        return self._send_query(QueryCmdId.GET_CAPABILITIES_VAL)

    def get_capabilities_names(self) -> GoProResp:
        """Get currently supported settings capabilities names.

        Raises:
            NotImplementedError: This isn't implemented on the camera
        """
        raise NotImplementedError("Not implemented on camera!")

    def register_value_update(self) -> GoProResp:
        """Register for asynchronous notifications when a given setting ID's value updates.

        Returns:
            GoProResp: Current value of respective setting ID
        """
        if (response := self._send_query(QueryCmdId.REG_SETTING_VAL_UPDATE)).is_ok:
            self._communicator._register_listener((QueryCmdId.SETTING_VAL_PUSH, self._identifier))
        return response

    def unregister_value_update(self) -> GoProResp:
        """Stop receiving notifications when a given setting ID's value updates.

        Returns:
            GoProResp: Status of unregister
        """
        if (response := self._send_query(QueryCmdId.UNREG_SETTING_VAL_UPDATE)).is_ok:
            self._communicator._unregister_listener((QueryCmdId.SETTING_VAL_PUSH, self._identifier))
        return response

    def register_capability_update(self) -> GoProResp:
        """Register for asynchronous notifications when a given setting ID's capabilities update.

        Returns:
            GoProResp: Current capabilities of respective setting ID
        """
        if (response := self._send_query(QueryCmdId.REG_CAPABILITIES_UPDATE)).is_ok:
            self._communicator._register_listener((QueryCmdId.SETTING_CAPABILITY_PUSH, self._identifier))
        return response

    def unregister_capability_update(self) -> GoProResp:
        """Stop receiving notifications when a given setting ID's capabilities change.

        Returns:
            GoProResp: Status of unregister
        """
        if (response := self._send_query(QueryCmdId.UNREG_CAPABILITIES_UPDATE)).is_ok:
            self._communicator._unregister_listener((QueryCmdId.SETTING_CAPABILITY_PUSH, self._identifier))
        return response

    def _build_cmd(self, cmd: QueryCmdId) -> bytearray:
        """Build the data to send a settings query over-the-air.

        Args:
            cmd (QueryCmdId): command to build

        Returns:
            bytearray: data to send over-the-air
        """
        ret = bytearray([cmd.value, self._identifier.value])
        return ret


class BleStatus(BleCommand[StatusId]):
    """An individual camera status that is interacted with via BLE."""

    UUID: Final[BleUUID] = GoProUUIDs.CQ_QUERY

    def __init__(self, communicator: GoProBle, identifier: StatusId, parser: QueryParserType) -> None:
        """Constructor

        Args:
            communicator (GoProBle): Adapter to read status data
            identifier (StatusId): ID of status
            parser (QueryParserType): construct to parse or enum to represent status value

        Raises:
            TypeError: Invalid parser type
        """
        if isinstance(parser, construct.Construct):
            parser_builder = construct_adapter_factory(parser)
        elif isinstance(parser, BytesParserBuilder):
            parser_builder = parser
        elif issubclass(parser, GoProEnum):
            parser_builder = enum_parser_factory(parser)
        else:
            raise TypeError(f"Unexpected {parser=}")
        BleCommand.__init__(self, communicator, uuid=self.UUID, parser=parser_builder, identifier=identifier)
        self._identifier = identifier

    def __str__(self) -> str:
        return str(self._identifier.name.lower().replace("_", " ")).title()

    def _send_query(self, response_id: QueryCmdId) -> GoProResp:
        """Build the byte data and query setting information

        Args:
            response_id (QueryCmdId): expected identifier of response

        Returns:
            GoProResp: query response
        """
        data = self._build_cmd(response_id)
        logger.info(
            Logger.build_log_tx_str(jsonify(self._as_dict(f"{response_id.name}.{self._identifier.name}")))
        )
        response = self._communicator._send_ble_command(self.UUID, data, response_id)
        logger.info(Logger.build_log_rx_str(response))
        return response

    def _as_dict(  # pylint: disable = arguments-differ
        self,
        identifier: Union[QueryCmdId, SettingId, str],
        *_: Any,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Return the attributes of the command as a dict

        Args:
            identifier (Union[QueryCmdId, SettingId, str]): identifier of the command for this send
            *_ (Any): unused
            **kwargs (Any): additional entries for the dict

        Returns:
            dict[str, Any]: command as dict
        """
        return dict(id=identifier, **self._base_dict) | kwargs

    def get_value(self) -> GoProResp:
        """Get the current value of a status.

        Returns:
            GoProResp: current status value
        """
        return self._send_query(QueryCmdId.GET_STATUS_VAL)

    def register_value_update(self) -> GoProResp:
        """Register for asynchronous notifications when a status changes.

        Returns:
            GoProResp: current status value
        """
        if (response := self._send_query(QueryCmdId.REG_STATUS_VAL_UPDATE)).is_ok:
            self._communicator._register_listener((QueryCmdId.STATUS_VAL_PUSH, self._identifier))
        return response

    def unregister_value_update(self) -> GoProResp:
        """Stop receiving notifications when status changes.

        Returns:
            GoProResp: Status of unregister
        """
        if (response := self._send_query(QueryCmdId.UNREG_STATUS_VAL_UPDATE)).is_ok:
            self._communicator._unregister_listener((QueryCmdId.STATUS_VAL_PUSH, self._identifier))
        return response

    def _build_cmd(self, cmd: QueryCmdId) -> bytearray:
        """Build the data for a given status command.

        Args:
            cmd (QueryCmdId): command to build data for

        Returns:
            bytearray: data to send over-the-air
        """
        ret = bytearray([cmd.value, self._identifier.value])
        return ret


######################################################## Wifi #################################################


class WifiGetJsonCommand(WifiCommand[str]):
    """A Wifi command that writes to a BleUUID (with parameters) and receives JSON as response"""

    def __init__(
        self,
        communicator: GoProWifi,
        endpoint: str,
        components: Optional[list[str]] = None,
        arguments: Optional[list[str]] = None,
        parser: Optional[JsonParser] = None,
        identifier: Optional[str] = None,
    ) -> None:
        if not identifier:
            # Build human-readable name from endpoint
            identifier = endpoint.lower().removeprefix("gopro/").replace("/", " ").replace("_", " ").title()
            try:
                identifier = identifier.split("?")[0].strip("{}")
            except IndexError:
                pass
        super().__init__(communicator, endpoint, identifier, components, arguments, parser)

    @no_type_check
    def __call__(self, **kwargs: Any) -> GoProResp:  # noqa: D102
        # Append components
        url = self._endpoint
        for component in self._components or []:
            url += "/" + kwargs.pop(component)
        # Append parameters
        if self._args:
            url += "?" + urlencode(
                {
                    k: kwargs[k].value if isinstance(kwargs[k], enum.Enum) else kwargs[k]
                    for k in self._args
                    if kwargs[k] is not None
                },
                safe="/",
            )
        # Send to camera
        logger.info(Logger.build_log_tx_str(jsonify(self._as_dict(**kwargs, endpoint=url))))
        response = self._communicator._get(url, self._parser)
        response._meta.append(self._identifier)
        logger.info(Logger.build_log_rx_str(response))
        return response


class WifiGetBinary(WifiCommand[str]):
    """A Wifi command that writes to a BleUUID (with parameters) and receives a binary stream as response"""

    def __init__(
        self,
        communicator: GoProWifi,
        endpoint: str,
        components: Optional[list[str]] = None,
        arguments: Optional[list[str]] = None,
        parser: Optional[JsonParser] = None,
        identifier: Optional[str] = None,
    ) -> None:
        if not identifier:
            # Build human-readable name from endpoint
            identifier = endpoint.lower().removeprefix("gopro/").replace("/", " ").replace("_", " ").title()
            try:
                identifier = identifier.split("?")[0].strip("{}")
            except IndexError:
                pass
        super().__init__(communicator, endpoint, identifier, components, arguments, parser)

    def __call__(self, /, camera_file: str, local_file: Optional[Path] = None) -> Path:  # noqa: D102
        # The method that will actually send the command and receive the stream
        local_file = local_file or Path(".") / camera_file
        url = self._endpoint + "/" + camera_file
        logger.info(
            Logger.build_log_tx_str(
                jsonify(self._as_dict(endpoint=url, camera_file=camera_file, local_file=local_file))
            )
        )
        # Send to camera
        self._communicator._stream_to_file(url, local_file)
        logger.info(
            Logger.build_log_rx_str(
                jsonify(self._as_dict(status="SUCCESS", endpoint=url, local_file=local_file))
            )
        )
        return local_file


class WifiSetting(WifiCommand[SettingId], Generic[ValueType]):
    """An individual camera setting that is interacted with via Wifi."""

    def __init__(self, communicator: GoProWifi, identifier: SettingId) -> None:
        super().__init__(
            communicator,
            "gopro/camera/setting?setting={}&option={}",
            identifier=identifier,
        )
        # Note! It is assumed that BLE and WiFi settings are symmetric so we only add to the communicator's
        # parser in the BLE Setting.

    def __str__(self) -> str:
        return str(self._identifier.name.lower().replace("_", " ")).title()

    def set(self, value: ValueType) -> GoProResp:
        """Set the value of the setting.

        Args:
            value (ValueType): value to set setting

        Returns:
            GoProResp: Status of set
        """
        url = self._endpoint.format(self._identifier.value, value)
        logger.info(Logger.build_log_tx_str(jsonify(self._as_dict(value=value, endpoint=url))))
        value = value.value if isinstance(value, enum.Enum) else value
        # Send to camera
        if response := self._communicator._get(url):
            response._meta.append(self._identifier)
            logger.info(Logger.build_log_rx_str(response))
        return response
